"""
Provides various types of plots and plotting utilities
"""

from .Graphics import Graphics, Graphics3D, GraphicsGrid
from .Backends import GraphicsAxes, GraphicsFigure
import numpy as np
# import matplotlib.figure
# import matplotlib.axes

__all__ = [
    "Plot", "DataPlot", "ArrayPlot", "TensorPlot",
    "Plot2D", "ListPlot2D",
    "Plot3D", "ListPlot3D",
    "CompositePlot",
]

######################################################################################################
#
#                                    'adaptive' function sampling
#
#
# region function application
def _apply_f(f, grid):
    try:
        vals = f(grid)
    except:
        vals = np.vectorize(f)(grid)

    return vals

def _semi_adaptive_sample_func(f, xmin, xmax, npts=150, max_refines=10, der_cut=10**5):

    refines = 0
    der_good = False

    try:
        from ..Zachary import finite_difference
    except ImportError:
        pass
    else:
        while refines < max_refines and not der_good:
            grid = np.linspace(xmin, xmax, npts)
            vals = _apply_f(f, grid)
            ders = finite_difference(grid, vals, 1, 5)
            der_good = not np.any(abs(ders) > der_cut)
            npts *= 2
            refines += 1

    if refines == 0:  # edge case
        grid = np.linspace(xmin, xmax, npts)
        vals = _apply_f(f, grid)

    return grid, vals, npts, refines

def _semi_adaptive_sample_func2(f, xmin, xmax, ymin, ymax, npts=15, max_refines=10, der_cut=10**5):
    from ..Zachary import finite_difference

    refines = 0
    der_good = False
    try:
        from ..Zachary import finite_difference
    except ImportError:
        pass
    else:
        while refines < max_refines and not der_good:
            grid = np.array(np.meshgrid(np.linspace(xmin, xmax, npts), np.linspace(ymin, ymax, npts))).T
            vals = _apply_f(f, grid)
            ders = finite_difference(grid, vals, (1, 1), (5, 5))
            der_good = not np.any(abs(ders) > der_cut)
            npts *= 2
            refines += 1

    if refines == 0:  # edge case
        grid = np.linspace(xmin, xmax, npts)
        vals = _apply_f(f, grid)

    return grid, vals, npts, refines
# endregion


######################################################################################################
#
#                                    Plot data methods
#
#
# region plot data prep
def _interp2DData(gpts, **opts):
    from scipy.interpolate import griddata

    x = np.sort(gpts[:, 0])
    y = np.sort(gpts[:, 1])

    xmin = np.min(x); xmax = np.max(x)

    xdiffs = np.abs(np.diff(x)); xh = np.min(xdiffs[np.nonzero(xdiffs)])
    ymin = np.min(y); ymax = np.max(y)
    ydiffs = np.abs(np.diff(y)); yh = np.min(xdiffs[np.nonzero(ydiffs)])

    num_x = (xmin - xmax) / xh
    if 5 > num_x or num_x < 10000:  # don't want to get too wild
        num_x = 100  # okay but let's get a little wild
    num_y = (ymin - ymax) / yh
    if 5 > num_y or num_y < 10000:  # don't want to get too wild
        num_y = 100  # okay but let's get a little wild

    # import sys
    # print(num_x, num_y, file = sys.stderr)

    xmesh = np.linspace(xmin, xmax, num_x); ymesh = np.linspace(ymin, ymax, num_y)
    xmesh, ymesh = np.meshgrid(xmesh, ymesh)
    mesh = np.array((xmesh, ymesh)).T
    vals = griddata(gpts[:, (0, 1)], gpts[:, 2], mesh, **opts)

    return xmesh, ymesh, vals.T

def _get_2D_plotdata(func, xrange):
    if hasattr(func, 'subs'):
        from sympy import lambdify
        sym, xrange = xrange
        xrange = np.arange(*xrange)
        fvalues = lambdify([sym], func)(xrange)
    elif not callable(func):
        fvalues = xrange
        xrange = func
    else:
        if len(xrange) == 3 and abs(xrange[2]) < abs(xrange[1] - xrange[0]):
            xrange = np.arange(*xrange)
            fvalues = _apply_f(func, xrange)
        elif len(xrange) > 2:
            fvalues = _apply_f(func, xrange)
        else:
            res = _semi_adaptive_sample_func(func, *xrange)
            xrange = res[0]
            fvalues = res[1]
    return xrange, fvalues

def _get_3D_plotdata(func, xrange, yrange):
    if not callable(func):
        fvalues = yrange
        yrange = xrange
        xrange = func
    else:
        if len(xrange) == 3 and abs(xrange[2]) < abs(xrange[1] - xrange[0]):
            xrange = np.arange(*xrange)
            yrange = np.arange(*yrange)
            xrange, yrange = mesh = np.meshgrid(xrange, yrange)
            fvalues = _apply_f(func, mesh)
        elif len(xrange) > 2 and len(yrange) > 2:
            xrange, yrange = mesh = np.meshgrid(xrange, yrange)
            fvalues = _apply_f(func, mesh)
        else:
            res = _semi_adaptive_sample_func2(func, *xrange, *yrange)
            xrange, yrange = res[0].T
            fvalues = res[1]

    return xrange, yrange, fvalues
# endregion

######################################################################################################
#
#                                    Unified PlotBase class
#
#
# Never implemented this

######################################################################################################
#
#                                    2D Plots on 2D Axes
#
#
class Plot(Graphics):
    """
    The base plotting class to interface into matplotlib or (someday 3D) VTK.
    In the future hopefully we'll be able to make a general-purpose `PlottingBackend` class that doesn't need to be `matplotlib` .
    Builds off of the `Graphics` class to make a unified and convenient interface to generating plots.
    Some sophisticated legwork unfortunately has to be done vis-a-vis tracking constructed lines and other plotting artefacts,
    since `matplotlib` is designed to infuriate.
    """

    line_params = {
        "linewidth", "linestyle", "color", "marker", "markersize",
        "markeredgewidth", "markeredgecolor", "markerfacecolor", "markerfacecoloralt",
        "fillstyle", "antialiased", "dash_capstyle", "solid_capstyle",
        "dash_joinstyle", "solid_joinstyle", "pickradius", "drawstyle", "markevery",
        'gid', "zorder", 'label'
    }
    patch_parms = {
        "agg_filter", "alpha", "animated", "antialiased", "capstyle",
        "clip_box", "clip_on", "clip_path", "color", "edgecolor", "facecolor",
        "figure", "fill", "gid", "hatch", "in_layout", "joinstyle",
         "label", "linestyle", "linewidth", "path_effects",
        "picker", "rasterized", "sketch_params", "snap",
        "transform", "url", "visible", "zorder"
    }

    opt_keys = Graphics.opt_keys | {"plot_style"}

    default_plot_style = {}
    style_mapping = {"format":"fmt"}
    known_styles = {"fmt"} | line_params
    method = "plot"
    def __init__(self,
                 *params,
                 method=None,
                 figure=None, axes=None, subplot_kw=None,
                 plot_style=None, theme=None,
                 **opts
                 ):
        """
        :param params: _empty_ or _x_, _y_ arrays or _function_, _xrange_
        :type params:
        :param plot_style: the plot styling options to be fed into the plot method
        :type plot_style: dict | None
        :param method: the method name as a string or functional form of the method to plot
        :type method: str | function
        :param figure: the Graphics object on which to plot (None means make a new one)
        :type figure: Graphics | None
        :param axes: the axes on which to plot (used in constructing a Graphics, None means make a new one)
        :type axes: None
        :param subplot_kw: the keywords to pass on when initializing the plot
        :type subplot_kw: dict | None
        :param colorbar: whether to use a colorbar or what options to pass to the colorbar
        :type colorbar: None | bool | dict
        :param opts: options to be fed in when initializing the Graphics
        :type opts:
        """

        self.graphics = None

        # we're gonna set things up so that we can have delayed evaluation of the plotting.
        # i.e. a Plot can be initialized but then do all its plotting later
        if plot_style is None:
            plot_style = {}
        for k,v in self.style_mapping.items():
            if k in opts:
                opts[v] = opts[k]
                del opts[k]
        for k in self.known_styles:
            if k in opts:
                plot_style[k] = opts[k]
                del opts[k]
        for k in self.default_plot_style:
            if k not in plot_style:
                plot_style[k] = self.default_plot_style[k]
        self._plot_style = plot_style
        self.plot_opts = opts
        self._initialized = False
        self._data = None

        super().__init__(figure=figure, axes=axes, theme=theme, subplot_kw=subplot_kw, **opts)
        self._init_opts['plot_style'] = plot_style
        if method is None:
           method = self.method
        if isinstance(method, str):
            method = self.axes.get_plotter(method)
        self._method = method

        if len(params) > 0:
            self.plot(*params)

    known_keys = Graphics.known_keys | {
        'method',
        'plot_style',
        'insert_default_styles'
    }
    @classmethod
    def filter_options(cls, opts, allowed=None):
        new = {}
        if allowed is None:
            allowed = cls.known_styles | cls.known_keys
        for k in opts.keys() & allowed:
                new[k] = opts[k]
        return new
    def _check_opts(self, opts):
        diff = opts.keys() - (self.known_styles | self.known_keys)
        if len(diff) > 0:
            raise ValueError("unknown options for {}: {}".format(
                type(self).__name__, list(diff)
            ))

    def _initialize(self):
        self._initialized = True
        self.set_options(**self.plot_opts)

    def _get_plot_data(self, func, xrange):
        xrange, fvalues = _get_2D_plotdata(func, xrange)
        return xrange, fvalues

    def _plot_data(self, *data, **plot_style):
        return self._method(*self._get_plot_data(*data), **plot_style)

    def plot(self, *params, insert_default_styles=True, **plot_style):
        """
        Plots a set of data & stores the result
        :return: the graphics that matplotlib made
        :rtype:
        """
        if insert_default_styles:
            plot_style = dict(self.plot_style, **plot_style)
        self._data = (params, plot_style)
        self.graphics = self._plot_data(*params, **plot_style)
        if not self._initialized:
            self._initialize()
        return self.graphics
    @property
    def artists(self):
        if self.graphics is None or isinstance(self.graphics, list):
            return self.graphics
        else:
            return [self.graphics]

    # def copy(self):
    #     return self.change_figure(None)
    def _change_figure(self, new, *init_args, **init_kwargs):
        """Creates a copy of the object with new axes and a new figure

        :return:
        :rtype:
        """
        # print(init_kwargs)
        # print(self._data[0], init_args)
        # print(init_kwargs, self._init_opts)
        return super()._change_figure(new, *self._data[0], *init_args, **init_kwargs)

    def clear(self):
        """
        Removes the plotted data
        """
        for g in self.graphics:
            self.axes.remove(g)
        self.graphics=None
    def restyle(self, **plot_style):
        """
        Replots the data with updated plot styling
        :param plot_style:
        :type plot_style:
        """
        self.clear()
        self.plot(*self.data, **plot_style)

    @property
    def data(self):
        """
        The data that we plotted
        """
        if self._data is None:
            raise ValueError("{} hasn't been plotted in the first place...")
        return self._data[0]
    @property
    def plot_style(self):
        """
        The styling options applied to the plot
        """
        if self._data is None:
            style = self._plot_style
        else:
            style = self._data[1]
        return style

    def add_colorbar(self, graphics=None, norm=None,  **kw):
        """
        Adds a colorbar to the plot
        """
        if self._initialized:
            if graphics is None and norm is None:
                graphics = self.graphics
            return super().add_colorbar(graphics=graphics, **kw)

    def set_graphics_properties(self, *which, **kw):
        if isinstance(self.graphics, tuple):
            for n,g in enumerate(self.graphics):
                if len(which) == 0 or n in which:
                    self.axes.set_graphics_properties(g, **kw)
        else:
            self.axes.set_graphics_properties(self.graphics, **kw)


    @classmethod
    def merge(cls, main, other, *rest, **kwargs):
        return CompositePlot(main, other, *rest, **kwargs)

    plot_classes = {}
    @classmethod
    def resolve_method(cls, mpl_name):
        return cls.plot_classes[mpl_name]
    # @classmethod
    # def merge_plots(cls, *plots, **styles):
    #     ...
    @classmethod
    def register(cls, plot_class):
        if plot_class.method in cls.plot_classes:
            cls.plot_classes[plot_class.__name__] = plot_class
        else:
            cls.plot_classes[plot_class.method] = plot_class
        return plot_class
Plot.register(Plot)

class CompositePlot:
    def __init__(self, main, other, *rest, **kwargs):
        self.kwargs = kwargs
        self.plots = [main, other, *rest]
    def merge(self, **kwargs):
        base = self.plots[0].change_figure(None, **kwargs)
        for p in self.plots[1:]:
            p.change_figure(base)
        return base
    def show(self, interactive=True):
        self._ref = self.merge(interactive=interactive, **self.kwargs)
        # self._ref.pyplot.mpl_connect()
        self._ref.show()
    def _ipython_display_(self):
        self.show()

@Plot.register
class FilledPlot(Plot):
    """
    Inherits from `Plot`.
    Plots a bunch of x values against a bunch of y values using the `scatter` method.
    """
    known_styles = { "where", "interpolate", "step", "data" } | Plot.patch_parms
    method = "fill_between"

@Plot.register
class ScatterPlot(Plot):
    """
    Inherits from `Plot`.
    Plots a bunch of x values against a bunch of y values using the `scatter` method.
    """
    known_styles = { "s", "c", "marker", "cmap", "norm", "vmin", "vmax", "alpha",
                     "linewidths", "edgecolors", "plotnonfinite", "data", "zorder"}
    style_mapping = {"color":"c", "marker_size":"s"}
    method = "scatter"

class ListScatterPlot(ScatterPlot):
    """
    Inherits from `Plot`.
    Plots a bunch of (x, y) points using the `scatter` method.
    """
    def __init__(self, griddata, **opts):
        super().__init__(griddata[:, 0], griddata[:, 1], **opts)

@Plot.register
class ErrorBarPlot(Plot):
    """
    Inherits from `Plot`.
    Plots error bars using the `errorbar` method.
    """
    known_styles = {"yerr", "xerr", "fmt", "ecolor", "elinewidth", "capsize", "barsabove",
    "lolims", "uplims", "xlolims", "xuplims", "errorevery", "capthick", "data"} | Plot.known_styles
    method = "errorbar"
class ListErrorBarPlot(ErrorBarPlot):
    """A Plot that pulls the errorbar data from a list"""
    def __init__(self, griddata, **opts):
        super().__init__(griddata[:, 0], griddata[:, 1], **opts)

@Plot.register
class StickPlot(Plot):
    """A Plot object that plots sticks"""

    default_plot_style = {'basefmt': " ", 'markerfmt': " "}
    known_styles = {"linefmt", "markerfmt", "basefmt", "bottom", "label", "orientation", "data", 'color', 'line_style'}
    method = "stem"
    def plot(self, *params, insert_default_styles=True, **plot_style):
        """
        Plots a set of data | stores the result
        :return: the graphics that matplotlib made
        :rtype:
        """
        if insert_default_styles:
            plot_style = dict(self.plot_style, **plot_style)
        # plot_style = dict(self.plot_style, **plot_style)
        if 'linewidth' in plot_style:
            lw = plot_style['linewidth']
            del plot_style['linewidth']
        else:
            lw = None
        lc = None
        if 'color' in plot_style:
            if 'linefmt' in plot_style:
                raise ValueError("modifying linefmt not currently supported")
            lc = plot_style['color']
            del plot_style['color']
        linefmt = ''
        ls = plot_style.pop('line_style', None)
        if ls is not None:
            if ls == 'dashed':
                if 'linefmt' in plot_style:
                    raise ValueError("modifying passed linefmt not currently supported")
                linefmt+="--"
            elif ls == 'dotted':
                if 'linefmt' in plot_style:
                    raise ValueError("modifying passed linefmt not currently supported")
                linefmt+="-."
            plot_style['linefmt'] = linefmt
        super().plot(*params, insert_default_styles=False, **plot_style)
        if lw is not None:
            self.set_graphics_properties(1, linewidth=lw)
        if lc is not None:
            self.set_graphics_properties(1, color=lc)
        return self.graphics
class ListStickPlot(StickPlot):
    """A Plot object that plots sticks from a list"""
    def __init__(self, griddata, **opts):
        super().__init__(griddata[:, 0], griddata[:, 1], **opts)

@Plot.register
class DatePlot(Plot):
    method = 'plot_date'
    known_styles = {'fmt', 'tz', 'xdate', 'ydate', 'data'} | Plot.known_styles
@Plot.register
class StepPlot(Plot):
    method = 'step'
    known_styles = {'where', 'data'} | Plot.known_styles
@Plot.register
class LogLogPlot(Plot):
    method = 'loglog'
    # known_styles = {}
@Plot.register
class SemiLogXPlot(Plot):
    method = 'semilogx'
    # known_styles = {}
@Plot.register
class SemilogYPlot(Plot):
    method = 'semilogy'
    # known_styles = {}
@Plot.register
class HorizontalFilledPlot(Plot):
    method = 'fill_betweenx'
    known_styles = {'where', 'step', 'interpolate', 'data'} | Plot.patch_parms
@Plot.register
class BarPlot(Plot):
    method = 'bar'
    known_styles = {'height', 'width', 'bottom', 'align', 'data'} | Plot.patch_parms
@Plot.register
class HorizontalBarPlot(Plot):
    method = 'barh'
    known_styles = {'width', 'height', 'left', 'align'} | Plot.patch_parms
# @Plot.register
# class BarLabelPlot(Plot):
#     method = 'bar_label'
#     known_styles = {'container', 'labels', 'fmt', 'label_type', 'padding'}  | Plot.patch_parms
@Plot.register
class EventPlot(Plot):
    method = 'eventplot'
    known_styles = {'positions', 'orientation', 'lineoffsets', 'linelengths', 'linewidths', 'colors', 'linestyles', 'data'} | Plot.line_params
@Plot.register
class PiePlot(Plot):
    method = 'pie'
    known_styles = {'explode', 'labels', 'colors', 'autopct', 'pctdistance',
                    'shadow', 'labeldistance', 'startangle', 'radius', 'counterclock',
                    'wedgeprops', 'textprops', 'center', 'frame', 'rotatelabels', 'normalize',
                    'data'} | Plot.patch_parms
@Plot.register
class StackPlot(Plot):
    method = 'stackplot'
    known_styles = {'labels', 'colors', 'baseline', 'data'} | Plot.patch_parms
@Plot.register
class BrokenHorizontalBarPlot(Plot):
    method = 'broken_barh'
    known_styles = {'xranges', 'yrange', 'data'} | Plot.patch_parms
@Plot.register
class VerticalLinePlot(Plot):
    """
    Plots a bunch of vertical lines
    """
    known_styles = {'ymin', 'ymax', 'colors', 'linestyles', 'label', 'data'} | Plot.line_params
    method = 'vlines'
    def _get_plot_data(self, x, y=1.0):
        if isinstance(y, (int, float)):
            y = [0, y]
        return (x, y)
    def _plot_data(self, *data, **plot_style):
        x, y = data
        return self._method(x, *y, **plot_style)
@Plot.register
class HorizontalLinePlot(Plot):
    """
    Plots a bunch of vertical lines
    """
    known_styles = {'xmin', 'xmax', 'colors', 'linestyles', 'label', 'data'} | Plot.line_params
    method = 'hlines'
    def _get_plot_data(self, y, x=None):
        if x is not None:
            x, y = y, x
        else:
            x = 1.0
        if isinstance(x, (int, float)):
            x = [0, x]
        return (x, y)
    def _plot_data(self, *data, **plot_style):
        x, y = data
        return self._method(y, *x, **plot_style)
#     known_styles = {'xmin', 'xmax', 'colors', 'linestyles', 'label', 'data'}
@Plot.register
class PolygonPlot(Plot):
    method = 'fill'
    known_styles = {'data'} | Plot.patch_parms

@Plot.register
class AxisHorizontalLinePlot(Plot):
    method = 'axhline'
    known_styles = {'xmin', 'xmax'} | Plot.line_params
@Plot.register
class AxisHorizontalSpanPlot(Plot):
    method = 'axhspan'
    known_styles = {'ymin', 'ymax', 'xmin', 'xmax'} | Plot.patch_parms
@Plot.register
class AxisVerticalLinePlot(Plot):
    method = 'axvline'
    known_styles = {'ymin', 'ymax'} | Plot.line_params
@Plot.register
class AxisVeticalSpanPlot(Plot):
    method = 'axvspan'
    known_styles = {'xmin', 'xmax', 'ymin', 'ymax'} | Plot.patch_parms
@Plot.register
class AxisLinePlot(Plot):
    method = 'axline'
    known_styles = {'xy1', 'xy2', 'slope'} | Plot.line_params

@Plot.register
class StairsPlot(Plot):
    method = 'stairs'
    known_styles = {'values', 'edges', 'orientation', 'baseline', 'fill', 'data'} | Plot.patch_parms

# class ClabelPlot(Plot):
#     method = 'clabel'
#     known_styles = {'CS', 'levels'}


# class AnnotatePlot(Plot):
#     method = 'annotate'
#     known_styles = {'text', 'xy'}
# class TextPlot(Plot):
#     method = 'text'
#     known_styles = {'s', 'fontdict'}
# class TablePlot(Plot):
#     method = 'table'
#     known_styles = {'cellText', 'cellColours', 'cellLoc', 'colWidths', 'rowLabels', 'rowColours', 'rowLoc', 'colLabels', 'colColours', 'colLoc',
#                     'loc', 'bbox', 'edges'}
# class ArrowPlot(Plot):
#     method = 'arrow'
#     known_styles = {'dx', 'dy'}
# class InsetAxesPlot(Plot):
#     method = 'inset_axes'
#     known_styles = {'bounds', 'transform', 'zorder'}
# class IndicateInsetPlot(Plot):
#     method = 'indicate_inset'
#     known_styles = {'bounds', 'inset_ax', 'transform', 'facecolor', 'edgecolor', 'alpha', 'zorder'}
# class IndicateInsetZoomPlot(Plot):
#     method = 'indicate_inset_zoom'
#     known_styles = {'inset_ax'}
# class SecondaryXaxisPlot(Plot):
#     method = 'secondary_xaxis'
#     known_styles = {'location', 'functions'}
# class SecondaryYaxisPlot(Plot):
#     method = 'secondary_yaxis'
#     known_styles = {'location', 'functions'}
# class BarbsPlot(Plot):
#     method = 'barbs'
#     known_styles = {'data'}


######################################################################################################
#
#                                    Pure Data Plots on 2D Axes
#
#
class DataPlot(Plot):
    """
    Makes a 2D plot of arbitrary data using a plot method that handles that data type
    """
    image_params = {'cmap', 'norm', 'aspect', 'interpolation', 'alpha', 'vmin', 'vmax', 'origin',
                    'extent', 'interpolation_stage', 'filternorm', 'filterrad', 'resample', 'url', 'data'}
    def __init__(self,
                 *params,
                 plot_style=None, method=None,
                 figure=None, axes=None, subplot_kw=None,
                 colorbar=None,
                 **opts
                 ):
        """
        :param params: _empty_ or _data_
        :type params:
        :param plot_style: the plot styling options to be fed into the plot method
        :type plot_style: dict | None
        :param method: the method name as a string
        :type method: str
        :param figure: the Graphics object on which to plot (None means make a new one)
        :type figure: Graphics | None
        :param axes: the axes on which to plot (used in constructing a Graphics, None means make a new one)
        :type axes: None
        :param subplot_kw: the keywords to pass on when initializing the plot
        :type subplot_kw: dict | None
        :param colorbar: whether to use a colorbar or what options to pass to the colorbar
        :type colorbar: None | bool | dict
        :param opts: options to be fed in when initializing the Graphics
        :type opts:
        """
        super().__init__(*params,
                         plot_style=plot_style, method=method,
                         colorbar=colorbar, figure=figure,
                         axes=axes, subplot_kw=subplot_kw,
                         **opts
                         )
    def _get_plot_data(self, data):
        return data,

@Plot.register
class HistogramPlot(DataPlot):
    """
    Makes a Histogram of data
    """
    method = 'hist'
    known_styles = {'bins', 'range', 'density', 'weights', 'cumulative',
                    'bottom', 'histtype', 'align', 'orientation', 'rwidth', 'log', 'color',
                    'label', 'stacked', 'data'}

@Plot.register
class HistogramPlot2D(DataPlot):
    """
    Makes a 2D histogram of data
    """
    method = 'hist2d'
    known_styles = {'bins', 'range', 'density', 'weights', 'cmin', 'cmax', 'data'}

@Plot.register
class SpectrogramPlot(DataPlot):
    method = 'specgram'
    known_styles = {'NFFT', 'Fs', 'Fc', 'detrend', 'window', 'noverlap',
                    'cmap', 'xextent', 'pad_to', 'sides', 'scale_by_freq', 'mode', 'scale', 'vmin',
                    'vmax', 'data'} | DataPlot.image_params
@Plot.register
class AutocorrelationPlot(DataPlot):
    method = 'acorr'
    known_styles = {'detrend', 'normed', 'usevlines', 'maxlags', 'linestyle', 'marker', 'data'} | Plot.known_styles
@Plot.register
class AngleSpectrumPlot(DataPlot):
    method = 'angle_spectrum'
    known_styles = {'Fs', 'Fc', 'window', 'pad_to', 'sides', 'data'}  | Plot.patch_parms
@Plot.register
class CoherencePlot(DataPlot):
    method = 'cohere'
    known_styles = {'NFFT', 'Fs', 'Fc', 'noverlap', 'pad_to', 'sides', 'scale_by_freq', 'data'} | Plot.known_styles
@Plot.register
class CrossSpectralDensityPlot(DataPlot):
    method = 'csd'
    known_styles = {'NFFT', 'Fs', 'Fc', 'detrend', 'window', 'noverlap', 'pad_to', 'sides', 'scale_by_freq', 'return_line', 'data'}  | Plot.known_styles
@Plot.register
class MagnitudeSpectrumPlot(DataPlot):
    method = 'magnitude_spectrum'
    known_styles = {'Fs', 'Fc', 'window', 'pad_to', 'sides', 'scale', 'data'} | Plot.known_styles
@Plot.register
class PhaseSpectrumPlot(DataPlot):
    method = 'phase_spectrum'
    known_styles = {'Fs', 'Fc', 'window', 'pad_to', 'sides', 'data'} | Plot.known_styles
@Plot.register
class PowerSpectralDensityPlot(DataPlot):
    method = 'psd'
    known_styles = {'NFFT', 'Fs', 'Fc', 'detrend', 'window', 'noverlap', 'pad_to', 'sides', 'scale_by_freq', 'return_line', 'data'} | Plot.known_styles
@Plot.register
class CrossCorrelationPlot(DataPlot):
    method = 'xcorr'
    known_styles = {'normed', 'usevlines', 'maxlags', 'data', 'linestyle', 'marker'} | Plot.known_styles
@Plot.register
class BoxPlot(DataPlot):
    method = 'boxplot'
    known_styles = {'notch', 'sym', 'vert', 'whis', 'positions', 'widths',
                    'patch_artist', 'bootstrap', 'usermedians', 'conf_intervals', 'meanline',
                    'showmeans', 'showcaps', 'showbox', 'showfliers', 'boxprops', 'labels',
                    'flierprops', 'medianprops', 'meanprops', 'capprops', 'whiskerprops',
                    'manage_ticks', 'autorange', 'zorder', 'data'}
@Plot.register
class ViolinPlot(DataPlot):
    method = 'violinplot'
    known_styles = {'dataset', 'positions', 'vert', 'widths', 'showmeans',
                    'showextrema', 'showmedians', 'quantiles', 'points', 'bw_method', 'data'}
# class ViolinPlot(Plot):
#     method = 'violin'
#     known_styles = {'vpstats', 'positions', 'vert', 'widths', 'showmeans',
#                     'showextrema', 'showmedians'}
@Plot.register
class BoxAndWhiskerPlot(DataPlot):
    method = 'bxp'
    known_styles = {'bxpstats', 'positions', 'widths', 'vert', 'patch_artist',
                    'shownotches', 'showmeans', 'showcaps', 'showbox', 'showfliers', 'boxprops',
                    'whiskerprops', 'flierprops', 'medianprops', 'capprops', 'meanprops',
                    'meanline', 'manage_ticks', 'zorder'}
@Plot.register
class HexagonalHistogramPlot(DataPlot):
    method = 'hexbin'
    known_styles = {'C', 'gridsize', 'bins', 'xscale', 'yscale', 'extent', 'cmap', 'norm', 'vmin',
                    'vmax', 'alpha', 'linewidths', 'edgecolors', 'mincnt', 'marginals', 'data', 'reduce_C_function'} | Plot.patch_parms

@Plot.register
@Plot.register
class QuiverPlot(DataPlot):
    method = 'quiver'
    known_styles = {"units", "angles", "scale", "scale_units", "width", "headwidth", "headlength",
                    "headaxislength", "minshaft", "minlength", "pivot", "color", "data"} | Plot.patch_parms
@Plot.register
class StreamPlot(DataPlot):
    method = 'streamplot'
    known_styles = {'density', 'linewidth', 'color', 'cmap', 'norm',
                    'arrowsize', 'arrowstyle', 'minlength', 'transform', 'zorder', 'start_points',
                    'maxlength', 'integration_direction', 'data'}

@Plot.register
class ArrayPlot(DataPlot):
    """
    Plots an array as an image
    """

    method = 'imshow'
    known_styles = DataPlot.image_params
    def __init__(self, *params,
                 plot_style=None, colorbar=None,
                 figure=None, axes=None, subplot_kw=None,
                 **opts
                 ):
        super().__init__(*params,
                         plot_style=plot_style,
                         colorbar=colorbar, figure=figure,
                         axes=axes, subplot_kw=subplot_kw,
                         **opts
                         )
    def _get_plot_data(self, data):
        if hasattr(data, 'toarray'):
            data = data.toarray()
        return data,
@Plot.register
class MatrixPlot(ArrayPlot):
    method = 'matshow'
@Plot.register
class SparsityPlot(ArrayPlot):
    method = 'spy'
    known_styles = {'precision', 'marker', 'markersize', 'aspect', 'origin'} | ArrayPlot.known_styles

class TensorPlot(GraphicsGrid):
    """
    Plots slices of a tensor as a grid
    """
    def __init__(self, tensor,
                 nrows=None, ncols=None,
                 plot_style=None, colorbar=None,
                 figure=None, axes=None, subplot_kw=None,
                 method='imshow', plot_class=None,
                 **opts
                 ):
        from operator import mul
        from functools import reduce
        tensor_shape = tensor.shape
        total_dim = reduce(mul, tensor_shape[:-2], 1)
        if nrows is None or ncols is None:
            if len(tensor_shape) == 3:
                nrows = 1
                ncols = tensor_shape[0]
            elif len(tensor_shape) == 4:  # best case
                nrows, ncols = tensor_shape[:2]
            else:
                if nrows is not None:
                    ncols = total_dim // nrows
                elif ncols is not None:
                    nrows = total_dim // ncols
                else:
                    ncols = 5
                    nrows = total_dim // ncols
        super().__init__(nrows=nrows, ncols=ncols,
                         figure=figure,
                         axes=axes,
                         subplot_kw=subplot_kw
                         )

        tensor = tensor.reshape((total_dim,) + tensor_shape[-2:])
        if plot_class is None:
            plot_class = lambda data,colorbar=colorbar,**opts:ArrayPlot(data, colorbar=colorbar,**opts)
        for i in range(nrows):
            for j in range(ncols):
                graphics = self.axes[i][j]
                self.axes[i][j] = plot_class(
                    tensor[nrows * i + j],
                    figure=graphics,
                    plot_style=plot_style,
                    method=method,
                    **opts
                )

######################################################################################################
#
#                                    3D Plots on 2D Axes
#
#

class Plot2D(Plot):
    """
    A base class for plots of 3D data but plotted on 2D axes
    """
    known_styles = {"corner_mask", "colors", "alpha", "cmap", "norm", "vmin", "vmax", "origin",
                    "extent", "locator", "extend", "xunits, yunits", "antialiased", "nchunk",
                    "linewidths", "linestyles", "hatches", "data"}
    method='contour'
    def __init__(self, *params,
                 plot_style=None,
                 colorbar=None,
                 figure=None,
                 axes=None,
                 subplot_kw=None,
                 **opts
                 ):
        """
        :param params: either _empty_ or _x_, _y_, _z_ arrays or _function_, _xrange_, _yrange_
        :type params:
        :param plot_style: the plot styling options to be fed into the plot method
        :type plot_style: dict | None
        :param method: the method name as a string
        :type method: str
        :param figure: the Graphics object on which to plot (None means make a new one)
        :type figure: Graphics | None
        :param axes: the axes on which to plot (used in constructing a Graphics, None means make a new one)
        :type axes: None
        :param subplot_kw: the keywords to pass on when initializing the plot
        :type subplot_kw: dict | None
        :param colorbar: whether to use a colorbar or what options to pass to the colorbar
        :type colorbar: None | bool | dict
        :param opts: options to be fed in when initializing the Graphics
        :type opts:
        """
        super().__init__(*params,
                         plot_style=plot_style,
                         colorbar=colorbar, figure=figure,
                         axes=axes, subplot_kw=subplot_kw,
                         **opts
                         )
    def _get_plot_data(self, func, xrange, yrange):
        return _get_3D_plotdata(func, xrange, yrange)
@Plot.register
class ContourPlot(Plot2D):
    method = 'contourf'
    known_styles = {"triangles", "mask", "levels", "colors",
                    "alpha", "cmap", "norm", "vmin", "vmax", "origin", "extent", "locator", "extend",
                    "xunits, yunits", "antialiased", "linewidths", "linestyles"}
@Plot.register
class DensityPlot(Plot2D):
    method = 'pcolormesh'
    known_styles = {'alpha', 'norm', 'cmap', 'vmin', 'vmax', 'shading', 'antialiased', 'data'} | Plot.patch_parms
@Plot.register
class HeatmapPlot(Plot2D):
    method = 'pcolor'
    known_styles = {'shading', 'alpha', 'norm', 'cmap', 'vmin', 'vmax', 'data'} | ArrayPlot.known_styles
@Plot.register
class TriPlot(Plot2D):
    """A Plot object that plots a triangulation bars"""
    method = 'triplot'
    known_styles = {"triangles", "mask"} | Plot.known_styles
class ListTriPlot(Plot2D):
    """A Plot that pulls the triangulation data from a list"""
    def __init__(self, griddata, **opts):
        super().__init__(griddata[:, 0], griddata[:, 1], **opts)
@Plot.register
class TriDensityPlot(Plot2D):
    method = 'tripcolor'
    known_styles = {'alpha', 'norm', 'cmap', 'vmin', 'vmax', 'shading', 'facecolors'}
@Plot.register
class TriContourLinesPlot(Plot2D):
    method = 'tricontour'
    known_styles = {"triangles", "mask", "levels", "colors",
                    "alpha", "cmap", "norm", "vmin", "vmax", "origin", "extent", "locator", "extend",
                    "xunits, yunits", "antialiased", "linewidths", "linestyles"}
@Plot.register
class TriContourPlot(Plot2D):
    method = 'tricontourf'
    known_styles = {"triangles", "mask", "levels", "colors",
                    "alpha", "cmap", "norm", "vmin", "vmax", "origin", "extent", "locator", "extend",
                    "xunits, yunits", "antialiased", "hatches"}

class ListPlot2D(Plot2D):
    """
    Convenience class that handles the interpolation first
    """
    def __init__(self,
                 *params,
                 plot_style=None,
                 method='contour',
                 colorbar=None,
                 figure=None,
                 axes=None,
                 subplot_kw=None,
                 interpolate=True,
                 **opts
                 ):
        """
        :param params: either _empty_ or and array of (_x_, _y_, _z_) points
        :type params:
        :param plot_style: the plot styling options to be fed into the plot method
        :type plot_style: dict | None
        :param method: the method name as a string
        :type method: str
        :param figure: the Graphics object on which to plot (None means make a new one)
        :type figure: Graphics | None
        :param axes: the axes on which to plot (used in constructing a Graphics, None means make a new one)
        :type axes: None
        :param subplot_kw: the keywords to pass on when initializing the plot
        :type subplot_kw: dict | None
        :param colorbar: whether to use a colorbar or what options to pass to the colorbar
        :type colorbar: None | bool | dict
        :param interpolate: whether to interpolate the data or not
        :type interpolate: bool
        :param opts: options to be fed in when initializing the Graphics
        :type opts:
        """
        self.interpolate = interpolate
        super().__init__(*params,
                         plot_style=plot_style, method=method,
                         colorbar=colorbar, figure=figure,
                         axes=axes, subplot_kw=subplot_kw,
                         **opts
                         )

    def _get_plot_data(self, *griddata, interpolate=None):
        if interpolate is None:
            interpolate = self.interpolate
        if len(griddata) == 3:
            x, y, z = griddata
        elif interpolate:
            x, y, z = _interp2DData(griddata[0])
        else:
            x = griddata[0][:, 0]
            y = griddata[0][:, 1]
            z = griddata[0][:, 2]

        return (x, y, z)

@Plot.register
class ListContourPlot(ContourPlot):
    _get_plot_data = ListPlot2D._get_plot_data
    def __init__(self,*params, interpolate=True, **opts):
        self.interpolate = interpolate
        super().__init__(*params, **opts)
@Plot.register
class ListDensityPlot(DensityPlot):
    _get_plot_data = ListPlot2D._get_plot_data
    def __init__(self,*params, interpolate=True, **opts):
        self.interpolate = interpolate
        super().__init__(*params, **opts)
@Plot.register
class ListTriContourPlot(TriContourPlot):
    _get_plot_data = ListPlot2D._get_plot_data
    def __init__(self,*params, interpolate=True, **opts):
        self.interpolate = interpolate
        super().__init__(*params, **opts)
@Plot.register
class ListTriDensityPlot(TriDensityPlot):
    _get_plot_data = ListPlot2D._get_plot_data
    def __init__(self,*params, interpolate=True, **opts):
        self.interpolate = interpolate
        super().__init__(*params, **opts)


######################################################################################################
#
#                                    3D Plots on 3D Axes
#
#
class Plot3D(Graphics3D):  # basically a mimic of the Plot class but inheriting from Graphics3D
    """A base class for 3D plots"""

    default_plot_style = {}
    style_mapping = {"format": "fmt"}
    known_styles = {"fmt"} | Plot.line_params
    method = 'plot_surface'
    def __init__(self, *params,
                 plot_style=None,
                 method=None, colorbar=None,
                 figure=None, axes=None, subplot_kw=None,
                 **opts
                 ):
        """
        :param params: either _empty_ or _x_, _y_, _z_ arrays or _function_, _xrange_, _yrange_
        :type params:
        :param plot_style: the plot styling options to be fed into the plot method
        :type plot_style: dict | None
        :param method: the method name as a string
        :type method: str
        :param figure: the Graphics object on which to plot (None means make a new one)
        :type figure: Graphics | None
        :param axes: the axes on which to plot (used in constructing a Graphics, None means make a new one)
        :type axes: None
        :param subplot_kw: the keywords to pass on when initializing the plot
        :type subplot_kw: dict | None
        :param colorbar: whether to use a colorbar or what options to pass to the colorbar
        :type colorbar: None | bool | dict
        :param opts: options to be fed in when initializing the Graphics
        :type opts:
        """

        super().__init__(figure=figure, axes=axes, subplot_kw=subplot_kw)
        if method is None:
           method = self.method
        if isinstance(method, str):
            method = getattr(self.axes, method)
        self._method = method

        # we're gonna set things up so that we can have delayed evaluation of the plotting.
        # i.e. a Plot3D can be initialized but then do all its plotting later
        if plot_style is None:
            plot_style = {}
        for k,v in self.style_mapping.items():
            if k in opts:
                opts[v] = opts[k]
                del opts[k]
        for k in self.known_styles:
            if k in opts:
                plot_style[k] = opts[k]
        for k in self.default_plot_style:
            if k not in plot_style:
                plot_style[k] = self.default_plot_style[k]
        self._plot_style = plot_style
        self.plot_style = plot_style
        self.plot_opts = opts
        self._colorbar = colorbar
        self._initialized = False

        if len(params) > 0:
            self.plot(*params)

    def _initialize(self):
        self._initialized = True
        self.set_options(**self.plot_opts)
        if self.colorbar:
            self.add_colorbar()
        elif isinstance(self.colorbar, dict):
            self.add_colorbar(**self.colorbar)

    def _get_plot_data(self, func, xrange, yrange):
        return _get_3D_plotdata(func, xrange, yrange)

    def _plot_data(self, *data, **plot_style):
        return self._method(*self._get_plot_data(*data), **plot_style)

    def plot(self, *params, **plot_style):
        plot_style = dict(self.plot_style, **plot_style)
        self.graphics = self._plot_data(*params, **plot_style)
        if not self._initialized:
            self._initialize()
        return self.graphics
    def add_colorbar(self, **kw):
        if self._initialized:
            fig:GraphicsFigure = self.figure
            ax:GraphicsAxes = self.axes
            return fig.create_colorbar(self.graphics, **kw)
        else:
            self._colorbar = kw
    plot_classes = {}
    @classmethod
    def resolve_method(cls, mpl_name):
        return cls.plot_classes[mpl_name]
    # @classmethod
    # def merge_plots(cls, *plots, **styles):
    #     ...
    @classmethod
    def register(cls, plot_class):
        if plot_class.method in cls.plot_classes:
            cls.plot_classes[plot_class.__name__] = plot_class
        else:
            cls.plot_classes[plot_class.method] = plot_class
        return plot_class

@Plot3D.register
class ScatterPlot3D(Plot3D):
    """
    Creates a ScatterPlot of 3D data
    """
    method = 'scatter'

@Plot3D.register
class WireframePlot3D(Plot3D):
    """
    Creates a Wireframe mesh plot of 3D data
    """
    method = 'plot_wireframe'

@Plot3D.register
class ContourPlot3D(Plot3D):
    """
    Creates a 3D ContourPlot of 3D data
    """
    method = 'contourf'

class ListPlot3D(Plot3D):
    """
    Convenience 3D plotting class that handles the interpolation first
    """
    method = 'contour'
    def __init__(self,
                 *params,
                 plot_style=None,
                 method=None,
                 colorbar=None,
                 figure=None,
                 axes=None,
                 subplot_kw=None,
                 interpolate=True,
                 **opts
                 ):
        """
        :param params: either _empty_ or and array of (_x_, _y_, _z_) points
        :type params:
        :param plot_style: the plot styling options to be fed into the plot method
        :type plot_style: dict | None
        :param method: the method name as a string
        :type method: str
        :param figure: the Graphics object on which to plot (None means make a new one)
        :type figure: Graphics | None
        :param axes: the axes on which to plot (used in constructing a Graphics, None means make a new one)
        :type axes: None
        :param subplot_kw: the keywords to pass on when initializing the plot
        :type subplot_kw: dict | None
        :param colorbar: whether to use a colorbar or what options to pass to the colorbar
        :type colorbar: None | bool | dict
        :param interpolate: whether to interpolate the data or not
        :type interpolate: bool
        :param opts: options to be fed in when initializing the Graphics
        :type opts:
        """
        self.interpolate = interpolate
        super().__init__(*params,
                         plot_style=plot_style, method=method,
                         colorbar=colorbar, figure=figure,
                         axes=axes, subplot_kw=subplot_kw,
                         **opts
                         )

    def _get_plot_data(self, *griddata, interpolate=None):
        if interpolate is None:
            interpolate = self.interpolate
        if len(griddata) == 3:
            x, y, z = griddata
        elif interpolate:
            x, y, z = _interp2DData(griddata[0])
        else:
            x = griddata[0][:, 0]
            y = griddata[0][:, 1]
            z = griddata[0][:, 2]

        return (x, y, z)

@Plot3D.register
class ListTriPlot3D(ListPlot3D):
    """
    Creates a triangulated surface plot in 3D
    """
    method = 'plot_trisurf'
    default_plot_style = {}


# add classes to __all__
for c in Plot.plot_classes.values():
    if c.__name__ not in __all__:
        __all__.append(c.__name__)
for c in Plot3D.plot_classes.values():
    if c.__name__ not in __all__:
        __all__.append(c.__name__)


import collections
import enum, types, ast
import os, sys, tempfile, uuid
import subprocess, inspect
import threading
import weakref

from ..Scaffolding import Checkpointer

__all__ = [
    "ScriptContext",
    "ScriptRunner"
]

class Sentinels(enum.Enum):
    Missing = 'missing'

class ScriptContext:
    def __init__(self,
                 objects:dict,
                 imports=None,
                 script_dir=None,
                 script_id=None,
                 context_file=None,
                 autodelete=None,
                 path=None):
        if script_dir is None:
            script_dir = self.get_dir()
        self.dir = script_dir
        self.imps = [] if imports is None else list(imports)
        self.objs = objects
        if path is None:
            path = sys.path
        self.path = path
        if script_id is None:
            script_id = str(uuid.uuid4())
        if context_file is None:
            if autodelete is None: autodelete = True
            context_file = self.get_context_file(script_id)
        self.autodelete = autodelete
        self.context_file = context_file
        self._depth = 0

    @classmethod
    def get_dir(cls):
        return os.getcwd()


    context_file_name = 'context-'
    context_file_extension = '.hdf5'
    @classmethod
    def get_context_file(cls, script_id):
        with tempfile.NamedTemporaryFile(prefix=cls.context_file_name+str(script_id), suffix=cls.context_file_extension) as tf:
            return tf.name

    def populate_context_file(self):
        if len(self.objs) > 0:
            mods = {
                k:v for k,v in self.objs.items()
                if isinstance(v, types.ModuleType)
            }
            imports = [
                name if val.__name__ == name else ["", val.__name__, name]
                for name, val in mods.items()
            ]
            objs = {
                k: v for k, v in self.objs.items()
                if not isinstance(v, types.ModuleType)
            }
            if len(mods) == len(self.objs):
                context_file = None
            else:
                with Checkpointer.from_file(self.context_file) as chk:
                    chk["_context_names"] = list(objs.keys())
                    for name,val in self.objs.items():
                        if not isinstance(val, types.ModuleType):
                            chk[name] = val
                context_file = self.context_file
            return self.imps + imports, context_file
        elif len(self.imps) > 0:
            return self.imps, None
        else:
            return None

    @classmethod
    def load_context(cls, context_file):
        chk = Checkpointer.from_file(context_file)
        keys = [
            k.decode() if not isinstance(k, str) else k
            for k in chk["_context_names"]
        ]
        with chk:
            return {
                k:chk[k]
                for k in keys
            }

    @classmethod
    def path_loader_template(cls, path):
        return f"""
import sys
for p in {path}:
    if p not in sys.path: sys.path.append(p)
"""

    @classmethod
    def context_loader_template(cls, context_file):
        return f"""
import McUtils.Jupyter as interactive
g = globals()
for k,v in interactive.ScriptContext.load_context("{context_file}").items():
    g[k] = v
"""

    @classmethod
    def chdir_template(cls, dir):
        return f"""
import os
os.chdir("{dir}")
"""

    @classmethod
    def imports_template(cls, imports):
        imp_lines = []

        for imp in imports:
            if isinstance(imp, str):
                imp_lines.append(f"import {imp}")
            else:
                if len(imp) == 1:
                    imp = ["", imp, ""]
                elif len(imp) == 2:
                    imp = [imp[0], imp[1], ""]

                from_, import_, as_ = ["" if i is None else i for i in imp]
                if len(from_) == 0:
                    if len(as_) == 0:
                        imp_lines.append(f"import {import_}")
                    else:
                        imp_lines.append(f"import {import_} as {as_}")
                else:
                    if len(as_) == 0:
                        imp_lines.append(f"from {from_} import {import_}")
                    else:
                        imp_lines.append(f"from {from_} import {import_} as {as_}")
        return "\n".join(imp_lines)

    def __enter__(self):
        # self.populate_context_file()
        self._depth += 1
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._depth = max([0, self._depth-1])
        if self._depth == 0 and self.autodelete:
            try:
                os.remove(self.context_file)
            except FileNotFoundError:
                pass

    def create_context_loader(self):
        blocks = []
        if len(self.path) > 0:
            blocks.append(self.path_loader_template(self.path))
        if self.dir is not None:
            blocks.append(self.chdir_template(self.dir))

        context = self.populate_context_file()
        if context is not None:
            imports, chk = context
            if len(imports) > 0:
                blocks.append(self.imports_template(imports))
            if chk is not None:
                blocks.append(self.context_loader_template(chk))

        return "\n".join(blocks)

    @classmethod
    def _handle_import_name(cls, name):
        if isinstance(name, ast.Name):
            return [name]
        elif isinstance(name, ast.alias):
            return [name.name, name.asname]

    @classmethod
    def extract_names(cls, code):
        imports = []
        names = []

        queue = collections.deque()
        queue.extend(ast.iter_child_nodes(code))
        while queue:
            node = queue.pop()
            if isinstance(node, ast.Import):
                imports.extend(
                    cls._handle_import_name(name)
                    for name in node.names
                )
            elif isinstance(node, ast.ImportFrom):
                imports.extend(
                    [node.module] + cls._handle_import_name(name)
                    for name in node.names
                )
            elif isinstance(node, ast.Attribute):
                queue.append(node.value)
            elif isinstance(node, ast.Name):
                names.append(node.id)
            else:
                queue.extend(ast.iter_child_nodes(node))

        return imports, names

    @classmethod
    def find_globals(cls):
        for frame in inspect.stack(1):
            globs = frame.frame.f_globals
            if globs['__name__'] == '__main__':
                return globs
        else:
            return inspect.stack(1)[1].frame.f_globals

    @classmethod
    def from_script(cls,
                    script,
                    modules=None, globs=None,
                    objects=None,
                    imports=None,
                    script_dir=None,
                    script_id=None,
                    context_file=None,
                    autodelete=None,
                    path=None
                    ):
        if modules is None:
            modules = sys.modules
        if globs is None:
            globs = cls.find_globals()

        code = ast.parse(script, mode='exec')
        imps, names = cls.extract_names(code)

        if objects is None:
            objects = {
                name:globs[name]
                for name in names
                if name in globs
            }

            ded = []
            for k, v in objects.items():
                if isinstance(v, type) and v.__module__ in modules:
                    if k == v.__name__:
                        imps.append([v.__module__, k, ""])
                    else:
                        imps.append([v.__module__, v.__name__, k])
                    ded.append(k)
            for d in ded:
                del objects[d]

        if imports is None:
            imports = imps
        # imps = {
        #     name:modules[name]
        #     for name in names
        #     if name in modules
        # }

        return cls(
            objects,
            imports=imports,
            script_id=script_id,
            script_dir=script_dir,
            context_file=context_file,
            autodelete=autodelete,
            path=path
        )


class ScriptRunner:

    def __init__(self, context, script_name='script-', script_suffix='.py', python=None, autodelete=None):
        self.context = context
        self.python = self.get_python(python)
        self.prefix = script_name
        self.suffix = script_suffix
        self.autodelete = autodelete or (autodelete is None)

    @classmethod
    def get_python(cls, python):
        if python is None:
            python = sys.executable
        return python

    class Result:
        __slots__ = ["script", "process", "thread"]
        def __init__(self, script, process=None, thread=None):
            self.script = script
            self.process = process
            self.thread = thread
        def join(self, timeout=None):
            if self.thread is not None:
                self.thread.join(timeout=timeout)
        @property
        def result(self):
            self.join()
            return self.process.stdout.decode()


    def _run_py_subprocess(self, result, script, dry_run=False, runner=None):
        py_script = None
        try:
            with tempfile.NamedTemporaryFile(prefix=self.prefix, suffix=self.suffix, mode='w+', delete=False) as py_script:
                py_script.write(script)
                py_file = py_script.name
            if dry_run:
                return (self.python, py_file)
            proc = subprocess.run([self.python, py_file], capture_output=True)
        finally:
            if not dry_run and self.autodelete and py_script is not None:
                try:
                    os.remove(py_script.name)
                except FileNotFoundError:
                    pass
            if self.context is not None:
                self.context.__exit__(None, None, None)
        result.process = proc
        if proc.returncode > 0:
            raise Exception(proc.stderr.decode())
        return proc.stdout.decode()

    def prep_script(self, script):
        if self.context is not None:
            script = self.context.create_context_loader() + script
        return script


    def run_script(self, script, dry_run=False, background=True, interactive=False):
        if self.context is not None:
            self.context.__enter__()
        script = self.prep_script(script)
        result = self.Result(script)

        if dry_run:
            return self._run_py_subprocess(result, script, dry_run=dry_run)
        elif interactive:
            from .Apps import DelayedResult
            return DelayedResult(self._run_py_subprocess, result, script, parent=self)
        elif background:
            thread = threading.Thread(target=self._run_py_subprocess, args=(result, script,))
            result.thread = thread
            thread.start()
            return result
        else:
            return self._run_py_subprocess(result, script)

    @classmethod
    def run(cls,
            script,
            modules=None, globs=None,
            objects=None,
            imports=None,
            script_dir=None,
            script_id=None,
            context_file=None,
            autodelete=None,
            path=None,
            background=True,
            interactive=False,
            dry_run=False
            ):
        context = ScriptContext.from_script(
            script,
            modules=modules, globs=globs,
            objects=objects,
            imports=imports,
            script_dir=script_dir,
            script_id=script_id,
            context_file=context_file,
            autodelete=autodelete,
            path=path
        )

        return cls(context, autodelete=autodelete).run_script(script,
                                                              dry_run=dry_run,
                                                              background=background,
                                                              interactive=interactive
                                                              )



import pprint, collections, enum

from .. import Devutils as dev
from .. import Numputils as nput

__all__ = ["TreeWrapper", "tree_traversal", "tree_iter", "TreeSentinels"]

class TreeTraversalOrder(enum.Enum):
    BreadthFirst = 'bfs'
    DepthFirst = 'dfs'

class TreeCallOrder(enum.Enum):
    PreVisit = "pre"
    PostVisit = "post"
    PostChildren = "final"

class TreeSentinels(enum.Enum):
    Stop = "stop"
    Skip = "skip"

def _get_tree_children(tree):
    if hasattr(tree, 'keys'):
        return list(enumerate(tree.keys()))
    else:
        return list(enumerate(tree))
def _get_tree_item(tree, item):
    return tree[item[1]]
def tree_traversal(tree, callback,
                   root=None,
                   get_item=None,
                   get_children=None,
                   visited:set=None,
                   check_visited=None,
                   traversal_ordering='bfs',
                   call_order='post'
                   ):
    if get_children is None and get_item is None:
        get_children, get_item = _get_tree_children, _get_tree_item
    elif get_children is not None and get_item is None:
        raise ValueError("`get_children` must be implemented if `get_item` is provided")
    elif get_children is None and get_item is not None:
        raise ValueError("`get_item` must be implemented if `get_children` is provided")

    if root is dev.default:
        root = get_children(tree)[0]
    if root in visited:
        return

    if check_visited is None:
        check_visited = visited is not None
    if check_visited and visited is None:
        visited = set()

    queue = collections.deque([[None, root]])
    if isinstance(traversal_ordering, str):
        traversal_ordering = TreeTraversalOrder(traversal_ordering)
    if isinstance(call_order, str):
        call_order = TreeCallOrder(call_order)
    if traversal_ordering is traversal_ordering.BreadthFirst:
        pop = queue.popleft
        extend = queue.extend
    else:
        pop = queue.popleft
        extend = queue.extendleft

    while queue:
        parent, head = pop()

        if call_order == TreeCallOrder.PreVisit:
            res = callback(parent, head, visited)
            if res is not None:
                return res

        if check_visited:
            visited.add(head)

        if call_order == TreeCallOrder.PostVisit:
            res = callback(parent, head, visited)
            if res is not None:
                return res

        if check_visited:
            extend(
                [head, get_item(head, h)]
                for h in get_children(head)
                if h not in visited
            )
        else:
            extend(
                [head, get_item(head, h)]
                for h in get_children(head)
            )

        if call_order == TreeCallOrder.PostChildren:
            res = callback(parent, head, visited)
            if res is not None:
                return res

def tree_iter(tree,
              root=None,
              get_item=None,
              get_children=None,
              visited: set=None,
              check_visited=None,
              traversal_ordering='bfs'
              ):
    if get_children is None and get_item is None:
        get_children, get_item = _get_tree_children, _get_tree_item
    elif get_children is not None and get_item is None:
        raise ValueError("`get_children` must be implemented if `get_item` is provided")
    elif get_children is None and get_item is not None:
        raise ValueError("`get_item` must be implemented if `get_children` is provided")

    if root is dev.default:
        root = get_children(tree)[0]
    if root in visited:
        return

    if check_visited is None:
        check_visited = visited is not None
    if check_visited and visited is None:
        visited = set()

    queue = collections.deque([[None, root]])
    if isinstance(traversal_ordering, str):
        traversal_ordering = TreeTraversalOrder(traversal_ordering)
    if traversal_ordering is traversal_ordering.BreadthFirst:
        pop = queue.popleft
        extend = queue.extend
    else:
        pop = queue.popleft
        extend = queue.extendleft

    while queue:
        parent, head = pop()
        res = yield parent
        if res is TreeSentinels.Skip:
            continue
        elif res is TreeSentinels.Stop:
            break

        if check_visited:
            visited.add(head)

        if check_visited:
            extend(
                [head, get_item(head, h)]
                for h in get_children(head)
                if h not in visited
            )
        else:
            extend(
                [head, get_item(head, h)]
                for h in get_children(head)
            )

class TreeWrapper:
    def __init__(self, tree):
        self.tree = tree

    def __repr__(self):
        fmt_tree = pprint.pformat(self.tree)
        cls = type(self)
        return f"{cls.__name__}({fmt_tree})"
    def __len__(self):
        return len(self.tree)
    def __iter__(self):
        return iter(self.tree)
    def condense_subtrees(self):
        if hasattr(self.tree, 'keys') or not all(
            hasattr(t, 'keys') for t in self.tree
        ):
            return self
        else:
            new_tree = self.tree[0]
            for t in self.tree[1:]:
                new_tree = dev.merge_dicts(new_tree, t)
            return type(self)(new_tree)
    def keys(self):
        if hasattr(self.tree, 'keys'):
            return self.tree.keys()
        else:
            return None
    def values(self):
        if hasattr(self.tree, 'values'):
            return self.tree.values()
        else:
            return self.tree
    def find_subtree(self, key):
        if hasattr(self.tree, 'keys'):
            return self.__getitem__(key)
        else:
            if nput.is_atomic(key):
                key = [key]
            for k in key:
                for n,v in enumerate(self.tree):
                    if dev.is_dict_like(v) and k in v:
                        return v[k]
                    # elif v == k:
                    #     return v
    @classmethod
    def get_tree_item(cls, tree, item):
        t = tree
        if nput.is_atomic(item):
            item = [item]
        base_exception = None
        for k in item:
            if not isinstance(k, str) and hasattr(t, 'keys'):
                woof = []
                for n,v in enumerate(t.keys()):
                    woof.append(v)
                    if n >= k:
                        break
                else:
                    base_exception = IndexError("index {} not valid for subtree with keys {}".format(
                        k, t.keys()
                    ))
                    break
                k = woof[-1]
            try:
                t = t[k]
            except (IndexError, KeyError, TypeError) as e:
                base_exception = e
                break


        if base_exception is not None:
            raise IndexError(f"{item} not found in tree") from base_exception

        return t
    def __getitem__(self, item):
        return self.get_tree_item(self.tree, item)
    def bfs(self, callback, **opts):
        #TODO: support most sophsticated children/item indexing
        return tree_traversal(self.tree, callback, traversal_ordering='bfs', **opts)
    def dfs(self, callback, **opts):
        return tree_traversal(self.tree, callback, traversal_ordering='dfs', **opts)


import types, typing

__all__ = [
    "HTMLableType",
    "WidgetableType",
    "InterfaceElementType",
    "ElementType",
    "ClassListType",
    "AttrsType"
]

@typing.runtime_checkable
class HTMLableType(typing.Protocol):
    def to_tree(self):
        ...
@typing.runtime_checkable
class WidgetableType(typing.Protocol):
    def to_widget(self):
        ...
InterfaceElementBaseType = typing.Union[
    str,
    typing.Mapping,
    HTMLableType,
    WidgetableType
]
InterfaceElementType = typing.Union[
    InterfaceElementBaseType,
    typing.Tuple[InterfaceElementBaseType, typing.Mapping],
]
ElementType = typing.Union[InterfaceElementType, None, typing.Iterable[InterfaceElementType]]
ClassListType = typing.Union[None, str, typing.Iterable[str]]
AttrsType = typing.Union[None, typing.Mapping]



import numpy as np

__all__ = [
    "MoleculeGraphics"
]


class MoleculeGraphics:
    def __init__(self,
                 atoms,
                 coords,
                 bonds=None,

                 displacements=None,
                 displacement_range=(-1, 1),
                 displacement_steps=5,

                 name='Molecule',  # 'My special little molecule I love it and do not hate secretely hate it',
                 program='Python',  # 'WTF-who-keeps-this-info-these-days',
                 comment="",  # "I have nothing to say to you",
                 metadata=None,
                 **params):

        self.obj = self._load_nglview()(
            atoms,
            coords,
            bonds=bonds,
            displacements=displacements,
            displacement_range=displacement_range,
            displacement_steps=displacement_steps,

            name=name,  # 'My special little molecule I love it and do not hate secretely hate it',
            program=program,  # 'WTF-who-keeps-this-info-these-days',
            comment=comment,  # "I have nothing to say to you",
            metadata=metadata,
            **params
        )
        self._widg = None

    def to_widget(self):
        if self._widg is None:
            self._widg = self.obj.show()
        return self._widg
    def show(self):
        return self.to_widget()
    def _ipython_display_(self):
        from .JHTML import JupyterAPIs
        JupyterAPIs().display_api.display(self.show())

    @classmethod
    def _load_nglview(cls):
        try:
            import nglview
        except ImportError:
            class MoleculeGraphics:
                def __init__(self, *args, **kwargs):
                    raise ImportError("{} requires `nglview`".format(type(self).__name__))
        else:
            class MoleculeGraphics(nglview.Structure, nglview.Trajectory):
                misc_useless_structural_data_header = " 0     0  0  0  0  0  0999 V2000"
                def __init__(self,
                             atoms,
                             coords,
                             bonds=None,

                             displacements=None,
                             displacement_range=(-1, 1),
                             displacement_steps=5,

                             name='Molecule',  # 'My special little molecule I love it and do not hate secretely hate it',
                             program='Python',  # 'WTF-who-keeps-this-info-these-days',
                             comment="",  # "I have nothing to say to you",
                             metadata=None,
                             **params
                             ):

                    super().__init__()

                    self.ext = "sdf"
                    self.params = params
                    self.atoms = atoms
                    self.coords = np.asanyarray(coords)
                    self.bonds = [] if bonds is None else bonds
                    self.name = name
                    self.program = program
                    self.comment = comment
                    self.meta = metadata

                    self.dips = displacements
                    if displacements is None:
                        self.scales = [0.]
                    else:
                        base_scales = np.linspace(*displacement_range, displacement_steps)
                        self.scales = np.concatenate([
                            base_scales,
                            base_scales[-2:0:-1]
                        ])

                def convert_header(self, comment=None):
                    return "\n".join([
                        self.name,
                        "  " + self.program,
                        " " + self.comment + ("" if comment is None else comment)
                    ])

                def convert_counts_line(self):
                    return "{:>3.0f}{:>3.0f} {}".format(len(self.atoms), len(self.bonds),
                                                        self.misc_useless_structural_data_header)

                def convert_coordinate_block(self, coords):
                    return "\n".join(
                        " {0[0]:>9.5f} {0[1]:>9.5f} {0[2]:>9.5f} {1:<3} 0  0  0  0  0  0  0  0  0  0  0  0".format(
                            crd,
                            at
                        ) for crd, at in zip(coords, self.atoms)
                    )

                def convert_bond_block(self):
                    return "\n".join(
                        "{:>3.0f}{:>3.0f}{:>3.0f}  0  0  0  0".format(
                            b[0] + 1,
                            b[1] + 1,
                            b[2] if len(b) > 2 else 1
                        ) for b in self.bonds
                    )

                def get_single_structure_string(self, coords, comment=None):
                    return """
{header}
{counts}
{atoms}
{bonds}
M  END
{meta}
$$$$
                    """.format(
                        header=self.convert_header(comment=comment),
                        counts=self.convert_counts_line(),
                        atoms=self.convert_coordinate_block(coords),
                        bonds=self.convert_bond_block(),
                        meta="" if self.meta is None else self.meta
                    ).strip()

                def get_coordinates(self, index):
                    if self.coords.ndim == 3:
                        return self.coords[index]
                    else:
                        if self.dips is None and not isinstance(index, (int, np.integer)) and not index == 0:
                            raise ValueError("no Cartesian displacements passed...")
                        scales = self.scales[index]
                        arr = self.coords + self.dips * scales
                        return arr

                def get_substructure(self, idx):
                    import copy

                    new = copy.copy(self)
                    new.coords = self.get_coordinates(idx)
                    new.dips = None
                    new.scales = [0.]
                    new.comment += "<Structure {}>".format(idx)
                    return new

                def __getitem__(self, idx):
                    return self.get_substructure(idx)

                def __iter__(self):
                    for i in range(self.n_frames):
                        yield self.get_substructure(i)

                def get_structure_string(self):
                    if self.dips is None and self.coords.ndim == 2:
                        return self.get_single_structure_string(self.coords)
                    else:
                        return "\n".join(
                            self.get_single_structure_string(self.get_coordinates(i)) for i in range(self.n_frames)
                        )

                @property
                def n_frames(self):
                    if self.coords.ndim == 2:
                        return len(self.scales)
                    else:
                        return len(self.coords)

                # basically arguments to "add representation"
                default_theme = [
                    ['licorice'],
                    ['ball+stick', dict(selection='_H', aspect_ratio=2.5)],
                    ['ball+stick', dict(selection='not _H', aspect_ratio=3)]
                ]

                def show(self,
                         themes=None,
                         frame_size=('100%', 'vh'),
                         scale=1.2,
                         **opts):
                    """Basically a hack tower to make NGLView actually visualize molecules well"""
                    viewer = nglview.NGLWidget(self)
                    if themes is None:
                        themes = self.default_theme
                    if len(themes) > 0:
                        viewer.clear()
                        for t in themes:
                            if len(t) > 1:
                                viewer.add_representation(t[0], **t[1])
                            else:
                                viewer.add_representation(t[0])
                    # try:
                    #     player = viewer.player.widget_player
                    # except AttributeError:
                    #     pass
                    # else:
                    #     if player is not None:
                    #         # player = player.children[0]
                    #         player._playing = playing
                    #         player._repeat = repeat
                    viewer.display(**opts)

                    # this is a temporary hack to get a better window size
                    if frame_size is not None:
                        arg_str = [s if isinstance(s, str) else '{}px'.format(s) for s in frame_size]
                        viewer._remote_call(
                            "setSize",
                            target="Widget",
                            args=arg_str
                        )
                    if scale is not None:
                        viewer[0].set_scale(scale)
                    return viewer

        return MoleculeGraphics


"""
This lists the types of readers and things available to the GaussianLogReader
"""
import numpy as np

from ....Parsers import *
from collections import namedtuple, OrderedDict

########################################################################################################################
#
#                                           OrcaLogComponents
#
# region OrcaLogComponents
Components = OrderedDict()  # we'll register on this bit by bit
# each registration should look like:

# Components["Name"] = {
#     "description" : string, # used for docmenting what we have
#     "tag_start"   : start_tag, # starting delimeter for a block
#     "tag_end"     : end_tag, # ending delimiter for a block None means apply the parser upon tag_start
#     "parser"      : parser, # function that'll parse the returned list of blocks (for "List") or block (for "Single")
#     "mode"        : mode # "List" or "Single"
# }

########################################################################################################################
#
#                                           CartesianCoordinates
#

# region CartesianCoordinates

 # the region thing is just a PyCharm hack to collapse the boilerplate here... Could also have done 5000 files

cartesian_start_tag = FileStreamerTag(
    """CARTESIAN COORDINATES (ANGSTROEM)""",
    follow_ups="---------------------------------"
)
cartesian_end_tag = "\n\n"

CartParser = StringParser(
    Repeating(
        (
            Named(
                Word, "Atoms",
                prefix=Optional(Whitespace),
                suffix=Whitespace
            ),
            Named(
                Repeating(
                    Capturing(Number),
                    min=3,
                    max=3,
                    prefix=Optional(Whitespace),
                    joiner = Whitespace
                ),
                "Coordinates", handler=StringParser.array_handler(dtype=float)
            )
        ),
        suffix=Optional(Newline)
    )
)

def strip_recursive(at_list):
    return [
        s.strip() if isinstance(s, str) else strip_recursive(s)
        for s in at_list
    ]
OrcaCoords = namedtuple("OrcaCoords", ["atoms", "coords"])
def cartesian_coordinates_parser(strs):
    strss = "\n\n".join(strs)

    parse = CartParser.parse_all(strss)
    coords = OrcaCoords(
        strip_recursive(parse["Atoms"].array),
        parse["Coordinates"].array
    )

    return coords

Components["CartesianCoordinates"] = {
    "tag_start": cartesian_start_tag,
    "tag_end"  : cartesian_end_tag,
    "parser"   : cartesian_coordinates_parser,
    "mode"     : "List"
}


cartesian_au_start_tag = FileStreamerTag(
    """CARTESIAN COORDINATES (A.U.)""",
    follow_ups="X           Y           Z"
)
cartesian_end_tag = "\n\n"

CartAUParser = StringParser(
    Repeating(
        (
           Named(
                Integer, "AtNos",
                prefix=Optional(Whitespace),
                suffix=Whitespace
            ),
            Named(
                Word, "Atoms",
                prefix=Optional(Whitespace),
                suffix=Whitespace
            ),
           Named(
                Number, "ZA",
                prefix=Optional(Whitespace),
                suffix=Whitespace
            ),
           Named(
                Integer, "Frag",
                prefix=Optional(Whitespace),
                suffix=Whitespace
            ),
           Named(
                Number, "Mass",
                prefix=Optional(Whitespace),
                suffix=Whitespace
            ),
            Named(
                Repeating(
                    Capturing(Number),
                    min=3,
                    max=3,
                    prefix=Optional(Whitespace),
                    joiner = Whitespace
                ),
                "Coordinates", handler=StringParser.array_handler(dtype=float)
            )
        ),
        suffix=Optional(Newline)
    )
)

OrcaAUCoords = namedtuple("OrcaAUCoords", ["atoms", "masses", "coords"])
def cartesian_au_coordinates_parser(strs):
    strss = "\n\n".join(strs)

    parse = CartAUParser.parse_all(strss)
    coords = OrcaAUCoords(
        strip_recursive(parse["Atoms"].array),
        parse["Mass"].array,
        parse["Coordinates"].array
    )

    return coords

Components["CartesianAUCoordinates"] = {
    "tag_start": cartesian_au_start_tag,
    "tag_end"  : cartesian_end_tag,
    "parser"   : cartesian_au_coordinates_parser,
    "mode"     : "List"
}


freqs_start_tag = FileStreamerTag(
    """VIBRATIONAL FREQUENCIES""",
    follow_ups="-----------------------"
)
freqs_end_tag = "\n\n\n"

FreqsParser = StringParser(
    Repeating(
        (
            Named(
                Integer, "Mode",
            ),
            ":",
            Whitespace,
            Named(
                Number, "Freqs"
            )
        ),
        suffix=Optional(Newline)
    )
)

def freqs_parser(freq_str):

    parse = FreqsParser.parse_all(freq_str)
    # coords = (
    #     strip_recursive(parse["Atoms"].array),
    #     parse["Freqs"].array
    # )

    return parse["Freqs"].array.flatten()

Components["VibrationalFrequencies"] = {
    "tag_start": freqs_start_tag,
    "tag_end"  : freqs_end_tag,
    "parser"   : freqs_parser,
    "mode"     : "Single"
}

OrcaMatrixParser = StringParser(
    Repeating(
        (
            Whitespace,
            Integer,
            Whitespace,
            Repeating(
                Capturing(Number),
                min=1,
                max=6,
                prefix=Optional(Whitespace),
                joiner=Whitespace
            ),
        ),
        handler=StringParser.array_handler(dtype=float),
        suffix=Optional(Newline)
    )
)

def parse_orca_matrix(orca_mat):

    parse = np.concatenate(
        [
            p.value[:, 1:]
            for p in OrcaMatrixParser.parse_iter(orca_mat)
        ],
        axis=1
    )
    # coords = (
    #     strip_recursive(parse["Atoms"].array),
    #     parse["Freqs"].array
    # )

    return parse

nms_start_tag = FileStreamerTag(
    """NORMAL MODES""",
    follow_ups="------------"
)
nms_end_tag = "\n\n\n"

Components["NormalModes"] = {
    "tag_start": nms_start_tag,
    "tag_end"  : nms_end_tag,
    "parser"   : parse_orca_matrix,
    "mode"     : "Single"
}


########################################################################################################################
#
#                                           Defaults
#
# region Defaults
Defaults = (
)
# endregion

########################################################################################################################
#
#                                           Ordering
#
# region Ordering
# defines the ordering in the log file
glk = ( # this must be sorted by what appears when

)
list_type = { k:-1 for k in Components if Components[k]["mode"] == "List" }
Ordering = { k:i for i, k in enumerate([k for k in glk if k not in list_type]) }
Ordering.update(list_type)
del glk
del list_type
# endregion

__components__ = Components
__ordering__ = Ordering
__defaults__ = Defaults



import numpy as np

from ..JHTML import JHTML, HTML, HTMLWidgets
from ..Apps import WrapperComponent
import uuid

__all__ = [
    # "D3API",
    "D3"
]
__reload_hook__ = ["..JHTML", "..Apps"]

class D3API:
    _api_versions = {}
    @classmethod
    def load(cls, version='v5'):
        if version not in cls._api_versions:
            cls._api_versions[version] = JHTML.JavascriptAPI(
    d3_init="""
let model = this.model;
if (typeof(model) === 'undefined' || model === null){
    model = this;
}

if (typeof(model.d3) === 'undefined' || model.d3 === null) {
    return import("https://d3js.org/d3.""" + version + """.min.js")
      .then(() => (model.d3 = d3));
} else {
    return Promise.resolve();
}""",

    d3_call="""
let model = this.model;
if (typeof(model) === 'undefined' || model === null){
    model = this;
}

model.callHandler("d3_init").then(()=>{
    let d3 = model.d3;
    let methods = event.content['methods'];
    let args = event.content['args'];
    let debug = event.content['debug'];
    let el = this.el;
    if (typeof(el) === 'undefined' || el === null) {
        if (event.content.hasOwnProperty('id')) {
            el = "#" + event.content['id']
        }
    }

    if (typeof(el) === 'undefined' || el === null) {
        alert("no d3 object provided");
    } else {
        let svg = d3.select(el);
        if (debug) { console.log(el, svg); }
        return methods.reduce(
            (acc, method, index) => {
                if (method in svg) {
                    let fn = svg[method];
                    if (debug) { console.log(method, args[index], fn); }
                    let argl = [svg, ...args[index]];
                    svg = fn.call(...argl);
                    if (debug) { console.log(svg); }
                } else {
                    if (debug) { console.log("no attr", method); }
                }
            }, 
            []
        )
    }
})
""",
    d3_props="""
let model = this.model;
if (typeof(model) === 'undefined' || model === null){
    model = this;
}

model.callHandler("d3_init").then(()=>{
    let d3 = model.d3;
    let props = event.content['properties'];
    let debug = event.content['debug'];
    let el = this.el;
    if (typeof(el) === 'undefined' || el === null) {
        if (event.content.hasOwnProperty('id')) {
            el = "#" + event.content['id']
        }
    }

    if (typeof(el) === 'undefined' || el === null) {
        alert("no d3 object provided");
    } else {
        let svg = d3.select(el);
        if (debug) { console.log(el, svg); }
        let attrs = props.map(
            (prop) => svg.attr(prop)
        );
        this.send({type:"content", content:attrs});
    }
})
"""
)
        return cls._api_versions[version]

def short_uuid(len=6):
    return str(uuid.uuid4()).replace("-", "")[:len]

class D3:
    """
    Provides a namespace for encapsultating various D3 object types
    """

    class Frame(WrapperComponent):
        """
        Provides a frame in which to run D3 commands + access to the D3 handles
        """

        wrappers = dict(wrapper=JHTML.Div)
        theme = dict(wrapper={'cls': []}, item={'cls': []})

        def __init__(self, *elems, version='v5', id=None, javascript_handles=None, dynamic=None, **attrs):
            if javascript_handles is not None:
                raise ValueError(...)
            self._api = D3API.load(version)
            if id is None:
                id = "d3-frame-" + short_uuid()
            self.id = id
            elems = [self.canonicalize_element(e) for e in elems]
            self.svg = D3.SVG(*elems, frame=self, id=self.id, **attrs)
            self.handles = {e['id'] for e in elems}
            super().__init__(
                self.svg,
                javascript_handles=self._api,
                dynamic=dynamic
            )

        def canonicalize_element(self, e):
            if isinstance(e, (HTML.XMLElement, HTMLWidgets.WrappedHTMLElement)):
                try:
                    id = e['id']
                except KeyError:
                    id = None
                if id is None:
                    e['id'] = self.id + "-" + short_uuid(3)
            else:
                raise NotImplementedError("not sure how I should canonicalize to HTML yet...")
            return e

        def call(self, commands, id=None, debug=False, **call_kwargs):
            methods = [c[0] for c in commands]
            args = [c[1:] for c in commands]
            w = self.to_widget()
            return w.call(
                'd3_call',
                id=self.id if id is None else id,
                methods=methods,
                args=args,
                debug=debug,
                **call_kwargs
            )

        def props(self, props, id=None):
            w = self.to_widget()
            return w.call(
                'd3_props',
                id=self.id if id is None else id,
                properties=props,
                return_message="content",
                timeout=1,
                callback=lambda *_, **__: [_, __],
                poll_interval=.1
            )

        def select(self, id):
            return self.Selection(self, id)

        def attr(self, execute=True, id=None, **attrs):
            calls = [
                        ['attr', k, v]
                        for k, v in attrs.items() if k != "style"
                    ] + [
                        ['style', k, v]
                        for k, v in attrs.get('style', {}).items()
                    ]
            if execute:
                self.call(calls, id=id)
            else:
                return self.CallChain(self, calls, id=id)

        def set_id(self, id, parent_id=None):
            if parent_id is not None:
                self.handles.remove(parent_id)
                self.handles.add(id)
            return self.call([['attr', 'id', id]], id=parent_id)

        def append(self, shape, id=None, parent_id=None, debug=False, **attrs):
            if isinstance(shape, D3.D3Element):
                shape.frame = self
                shape.initialize()
                return shape
            elif isinstance(shape, str):
                if id is None:
                    id = self.id + "-" + short_uuid(3)
                self.handles.add(id)
                self.call(
                    [
                        ['append', shape],
                        ['attr', 'id', id],
                    ] + [
                        ['attr', k, v]
                        for k, v in attrs.items() if k != "style"
                    ] + [
                        ['style', k, v]
                        for k, v in attrs.get('style', {}).items()
                    ],
                    id=parent_id
                )
                return self.Selection(self, id)
            else:
                raise ValueError("can't append {}".format(shape))
        def insert(self, before_id, shape, id=None, parent_id=None, debug=False, **attrs):
            if before_id is None:
                return self.append(shape, id=id, parent_id=parent_id, debug=debug, **attrs)
            if isinstance(shape, D3.D3Element):
                shape.frame = self
                shape.insert_in_frame(where)
                return shape
            else:
                if id is None:
                    id = self.id + "-" + short_uuid(3)
                self.handles.add(id)
                self.call(
                    [
                        ['insert', shape, "#"+before_id],
                        ['attr', 'id', id],
                    ] + [
                        ['attr', k, v]
                        for k, v in attrs.items() if k != "style"
                    ] + [
                        ['style', k, v]
                        for k, v in attrs.get('style', {}).items()
                    ],
                    id=parent_id,
                    debug=debug
                )
                return self.Selection(self, id)
        def remove(self, id=None):
            if id is None:
                id = self.id
            if id in self.handles:
                self.handles.remove(id)
            self.call(
                [
                    ['remove']
                ],
                id=id
            )
        def transition(self, id=None):
            return self.CallChain(self, [['transition']], id=id)

        class CallChain:
            def __init__(self, frame, commands, id=None):
                self.frame = frame
                self.commands = commands
                self.id = id

            def __getattr__(self, command):
                return self.call(self, command)

            class call:
                def __init__(self, parent, command):
                    self.parent = parent
                    self.command = command

                def __call__(self, *args):
                    return type(self.parent)(
                        self.parent.frame,
                        self.parent.commands + [[self.command] + list(args)],
                        id=self.parent.id
                    )

            def execute(self):
                self.frame.call(
                    self.commands,
                    id=self.id
                )

        class Selection:
            def __init__(self, frame, handle):
                self.frame = frame
                self.handle = handle

            def call(self, commands):
                return self.frame.call(commands)

            def props(self, properties):
                return self.frame.props(properties, id=self.handle)

            def transition(self):
                return self.frame.transition(id=self.handle)

    _cls_map = None
    @classmethod
    def get_class_map(cls):
        if cls._cls_map is None:
            cls._cls_map = {}
            for v in cls.__dict__.values():
                if isinstance(v, type) and hasattr(v, 'tag'):
                    cls._cls_map[v.tag] = v
        return cls._cls_map

    class D3Element(HTML.XMLElement):
        """
        Provides a hook into the XMLElement API to provide a model
        that can talk with `d3` through a `D3.Frame`
        """

        def __init__(self, tag, *elems, frame:'D3.Frame'=None, id=None, on_update=None, style=None, activator=None, **attrs):
            self._frame = frame
            self._had_id = id is not None
            if id is None and self.frame is not None:
                id = self.frame.id + "-" + short_uuid(3)
            # self._id = id
            if on_update is None:
                on_update = self._on_update
            super().__init__(tag, id=id, on_update=on_update, style=style, **attrs)
            self._elems = [self._wrap_d3(e) for e in elems]
        def __repr__(self):
            return "{}<{}>".format(type(self).__name__, self.id)

        @property
        def frame(self):
            return self._frame
        @frame.setter
        def frame(self, frame):
            self.set_frame(frame)
        def set_frame(self, frame, parent_id=None):
            if frame is not self._frame:
                self._frame = frame
                if not self._had_id:
                    if frame is None:
                        self._attrs['id'] = None
                    elif parent_id is None:
                        self._attrs['id'] = frame.id + "-" + short_uuid(3)
                    else:
                        self._attrs['id'] = parent_id + "-" + short_uuid(3)
                for e in self.elems:
                    if not isinstance(e, (str, int, float, bool)):
                        e.set_frame(frame, parent_id=self.id)
        @property
        def id(self):
            return self._attrs.get('id', None)
        def set_id(self, parent_id, overwrite=False):
            if overwrite or not self._had_id:
                new_id = parent_id + "-" + short_uuid(3)
                if self.frame is None:
                    self._attrs["id"] = new_id # doesn't need to call
                else:
                    self['id'] = new_id # do call an update
            for k in self.elems:
                if not isinstance(k, (str, int, float, bool)):
                    k.set_id(self.id, overwrite=overwrite)

        def initialize(self, parent_id=None):
            tag, attrs = self.to_d3()
            self.frame.append(
                tag,
                **attrs,
                parent_id=parent_id
            )
            self.initialize_children()
        def initialize_children(self):
            for k in self.elems:
                if not isinstance(k, (str, int, float, bool)):
                    k.initialize(parent_id=self.id)

        def to_d3(self):
            style = self.style
            if style is None:
                style = {}
            return [self.tag, dict(self.attrs, id=self.id, style=style)]
        def __setitem__(self, item, value):
            if isinstance(value, (D3.D3Element, HTML.XMLElement)):
                value = self._wrap_d3(value)
            super().__setitem__(item, value)
        def insert(self, where, child):
            if isinstance(child, (D3.D3Element, HTML.XMLElement)):
                child = self._wrap_d3(child)
            super().insert(where, child)
        def set_elems(self, elems):
            super().set_elems([self._wrap_d3(e) for e in elems])
        @staticmethod
        def _on_update(element:'D3.D3Element', key, value, old_value, caller, subkey=None):
            if element.frame is not None:
                element.frame.invalidate_cache()
            if key == 'attributes':
                element.reset_attributes(value, old_value)
            elif key == 'elements':
                element.reset_children(value, old_value)
            elif key == 'attribute':
                element.set_attribute(subkey, value, old_value)
            elif key == 'element':
                if value is None:
                    element.remove_child(subkey)
                elif old_value is None:
                    element.insert_child(subkey, value)
                else:
                    element.replace_child(subkey, old_value, value)
            else:
                raise NotImplementedError("haven't implemented change {}".format(key))
        def reset_attributes(self, new_attrs, old_attrs):
            removed_attrs = old_attrs - new_attrs.keys()
            settable_attrs = dict(new_attrs, **{k:None for k in removed_attrs})
            if 'id' in settable_attrs:
                del settable_attrs['id']
            self.frame.attr(id=self.id, **settable_attrs)
        def set_attribute(self, attr, new_value, old_value):
            self.frame.attr(id=self.id, **{attr:new_value})
        def insert_child(self, where, new_value:'D3.D3Element'):
            tag, attrs = new_value.to_d3()
            if where is None or where >= len(self.elems):
                before_id = None
            else:
                before_id = self.elems[where].id
            if before_id is not None and before_id == new_value.id:
                # insert_child usually happens after insert is called on base data
                if where < len(self.elems) - 1:
                    before_id = self.elems[where] + 1
                else:
                    before_id = None
            self.frame.insert(
                before_id,
                tag,
                **attrs,
                parent_id=self.id
            )
            new_value.initialize_children()
        def replace_child(self, where, old, new:'D3.D3Element'):
            self.frame.remove(old.id)
            self.insert_child(where, new)
        def reset_children(self, new_children:'Iterable[D3.D3Element]', old_children:'Iterable[D3.D3Element]'):
            for k in old_children:
                self.frame.remove(k.id)
            for n in new_children:
                tag, attrs = n.to_d3()
                self.frame.append(
                    tag,
                    parent_id=self.id,
                    **attrs
                )

        def _wrap_d3(self, e)->'D3.D3Element':
            if isinstance(e, D3.D3Element):
                if e.id is None:
                    if self.id is not None:
                        e.set_id(self.id)
                    #     e._id = self.id + "-" + short_uuid(3)
                    #     print("-->", e._id)
                    # else:
                    #     print(e, self)
                e.frame = self.frame
            elif isinstance(e, HTML.XMLElement):
                cls = D3.get_class_map().get(e.tag, None)
                id = e.attrs.get('id', None)
                if id is None and self.id is not None:
                    id = self.id + "-" + short_uuid(3)
                if cls is None:
                    e = D3.D3Element(e.tag, *e.elems, id=id, frame=self.frame, on_update=e.on_update, **{k:v for k in e.attrs.items() if k!='id'})
                else:
                    e = cls(*e.elems, id=id, frame=self.frame, on_update=e.on_update, **{k:v for k in e.attrs.items() if k!='id'})
            elif isinstance(e, (str, int, float)):
                pass
            else:
                raise NotImplementedError("don't know how to handle {}".format(e))
            return e

    class TagElement(D3Element):
        tag = None
        def __init__(self, *elems, frame=None, **attrs):
            super().__init__(self.tag, *elems, frame=frame, **attrs)
        def __call__(self, *elems, **kwargs):
            return type(self)(
                self._elems + list(elems),
                frame=self.frame,
                activator=self.activator,
                on_update=self.on_update,
                **dict(self.attrs, **kwargs)
            )

    class A(TagElement): tag="a"
    Link = A
    class Animate(TagElement): tag="animate"
    class AnimateMotion(TagElement): tag="animateMotion"
    class AnimateTransform(TagElement): tag="animateTransform"
    class Circle(TagElement): tag="circle"
    class ClipPath(TagElement): tag="clipPath"
    class Defs(TagElement): tag="defs"
    Definitions = Defs
    class Desc(TagElement): tag="desc"
    Description = Desc
    class Ellipse(TagElement): tag="ellipse"
    class FeBlend(TagElement): tag="feBlend"
    FilterBlend = FeBlend
    class FeColorMatrix(TagElement): tag="feColorMatrix"
    FilterColorMatrix = FeColorMatrix
    class FeComponentTransfer(TagElement): tag="feComponentTransfer"
    FilterComponentTransfer = FeComponentTransfer
    class FeComposite(TagElement): tag="feComposite"
    FilterComposite = FeComposite
    class FeConvolveMatrix(TagElement): tag="feConvolveMatrix"
    FilterConvolveMatrix = FeConvolveMatrix
    class FeDiffuseLighting(TagElement): tag="feDiffuseLighting"
    FilterDiffuseLighting = FeDiffuseLighting
    class FeDisplacementMap(TagElement): tag="feDisplacementMap"
    FilterDisplacementMap = FeDisplacementMap
    class FeDistantLight(TagElement): tag="feDistantLight"
    FilterDistantLight = FeDistantLight
    class FeDropShadow(TagElement): tag="feDropShadow"
    FilterDropShadow = FeDropShadow
    class FeFlood(TagElement): tag="feFlood"
    FilterFlood = FeFlood
    class FeFuncA(TagElement): tag="feFuncA"
    FilterAlphaChannelFunction = FeFuncA
    class FeFuncB(TagElement): tag="feFuncB"
    FilterBlueChannelFunction = FeFuncB
    class FeFuncG(TagElement): tag="feFuncG"
    FilterGreenChannelFunction = FeFuncG
    class FeFuncR(TagElement): tag="feFuncR"
    FilterRedChannelFunction = FeFuncR
    class FeGaussianBlur(TagElement): tag="feGaussianBlur"
    FilterGaussianBlur = FeGaussianBlur
    class FeImage(TagElement): tag="feImage"
    FilterImage = FeImage
    class FeMerge(TagElement): tag="feMerge"
    FilterMerge = FeMerge
    class FeMergeNode(TagElement): tag="feMergeNode"
    FilterMergeNode = FeMergeNode
    class FeMorphology(TagElement): tag="feMorphology"
    FilterMorphology = FeMorphology
    class FeOffset(TagElement): tag="feOffset"
    FilterOffset = FeOffset
    class FePointLight(TagElement): tag="fePointLight"
    FilterPointLight = FePointLight
    class FeSpecularLighting(TagElement): tag="feSpecularLighting"
    FilterSpecularLighting = FeSpecularLighting
    class FeSpotLight(TagElement): tag="feSpotLight"
    FilterSpotLight = FeSpotLight
    class FeTile(TagElement): tag="feTile"
    FilterTile = FeTile
    class FeTurbulence(TagElement): tag="feTurbulence"
    FilterTurbulence = FeTurbulence
    class Filter(TagElement): tag="filter"
    class ForeignObject(TagElement): tag="foreignObject"
    class G(TagElement): tag="g"
    Group = G
    class Hatch(TagElement): tag="hatch"
    class HatchPath(TagElement): tag="hatchpath"
    class Image(TagElement): tag="image"
    class Line(TagElement): tag="line"
    class LinearGradient(TagElement): tag="linearGradient"
    class Marker(TagElement): tag="marker"
    class Mask(TagElement): tag="mask"
    class Metadata(TagElement): tag="metadata"
    class MPath(TagElement): tag="mpath"
    MotionPath = MPath
    class Path(TagElement): tag="path"
    class Pattern(TagElement): tag="pattern"
    class Polygon(TagElement): tag="polygon"
    class Polyline(TagElement): tag="polyline"
    class RadialGradient(TagElement): tag="radialGradient"
    class Rect(TagElement): tag="rect"
    class Script(TagElement): tag="script"
    class Set(TagElement): tag="set"
    class SolidColor(TagElement): tag="solidcolor"
    class Stop(TagElement): tag="stop"
    class Style(TagElement): tag="style"
    class SVG(TagElement): tag="svg"
    class Switch(TagElement): tag="switch"
    class Symbol(TagElement): tag="symbol"
    class Text(TagElement): tag="text"
    class TextPath(TagElement): tag="textPath"
    class Title(TagElement): tag="title"
    class Tspan(TagElement): tag="tspan"
    TextSpan = Tspan
    class Use(TagElement): tag="use"
    class View(TagElement): tag="view"

    class Plots:
        """
        Helper namespace for wrapping matplotlib plots
        """

        @classmethod
        def use_as_backend(cls):
            import matplotlib
            return matplotlib.use('module://' + cls.__module__ + '_backend')

        #TODO: encapsulate all of this in some sort of wrapper object
        @classmethod
        def get_plot_object(cls, figure):
            if hasattr(figure, 'figure'):
                figure = figure.figure
            return figure.canvas.manager.frame.svg

        @classmethod
        def render_mpl(cls, figure, mpl_objs):
            from .d3_backend import FigureCanvasD3
            return FigureCanvasD3.render_objects(figure, mpl_objs)

        @classmethod
        def get_mpl_plot_bounds(cls, figure):
            import McUtils.Parsers as parse

            svg = cls.get_plot_object(figure)
            # w, h = [float(x) for x in parse.Number.findall(svg.attrs['width'] + ' ' + svg.attrs['height'])]
            # [vmx, vmy, vMX, vMY] = [float(x) for x in parse.Number.findall(svg.attrs['viewBox'])]

            patch = svg.find(
                svg.build_selector(
                    dict(node_type='g', cls='axes'),
                    dict(node_type='g', cls='patch')
                )
            ).elems[0]

            mx, my, _, _, Mx, My, _, _ = [float(x) for x in parse.Number.findall(patch.attrs['d'])]
            return figure.plot_range, [[mx, my], [Mx, My]]

        @classmethod
        def to_plot_coords(cls, figure, data_points):
            ((dmx, dMx), (dmy, dMy)), ((mx, my), (Mx, My)) = cls.get_mpl_plot_bounds(figure)
            drx = dMx - dmx
            dry = dMy - dmy
            rx = Mx - mx
            ry = My - my
            data_points = np.asanyarray(data_points)
            smol = data_points.ndim == 1
            if smol:
                data_points = data_points[np.newaxis]
            scaled_x = (data_points[:, 0] - dmx) * rx / drx
            scaled_y = (data_points[:, 1] - dmy) * ry / dry

            data = np.array([scaled_x + mx, scaled_y + my]).T
            if smol:
                data = data[0]
            return data






        # class Wrapper:
        #     def __init__(self, figure):



"""
Provides utilities for setting up platform-independent parallelism
in a hopefully unobtrusive way.

This is used more extensively in `Psience`, but the design is to unify the MPI and `multiprocessing` APIs
so that one can simply pass in a `Parallelizer` object to a function and obtain parallelism over as
many processes as that object supports.
As a fallthrough, a `SerialNonParallelizer` is provided as a subclass that handles serial evaluation with
the same API so fewer special cases need to be checked.
Any function that supports parallelism should take the `parallelizer` keyword, which will be fed
the `Parallelizer` object itself.
"""

__all__ = []
from .Parallelizers import *; from .Parallelizers import __all__ as exposed
__all__ += exposed
from .Runner import *; from .Runner import __all__ as exposed
__all__ += exposed
from .SharedMemory import *; from .SharedMemory import __all__ as exposed
__all__ += exposed

def _ipython_pinfo_():
    from ..Docs import jdoc
    import sys

    return jdoc(sys.modules[__name__])



from .BaseSurface import *
import numpy as np
from collections import namedtuple

__all__ = [
    "Surface",
    "MultiSurface"
]

class Surface:
    """
    This actually isn't a concrete implementation of BaseSurface.
    Instead it's a class that _dispatches_ to an implementation of BaseSurface to do its core evaluations (plus it does shape checking)
    """
    def __init__(self, data, dimension=None, base=None, **metadata):
        """

        :param data:
        :type data:
        :param dimension:
        :type dimension:
        :param base:
        :type base: None | Type[BaseSurface]
        :param metadata:
        :type metadata:
        """
        self.metadata = metadata
        if len(data) == 2 and isinstance(data[1], dict):
            data, opts = data
        elif isinstance(data, dict):
            opts = data
            data = ()
        else:
            opts = {}
        if base is None:
            base = self.detect_base(data, opts)
        self.base = base(*data, dimension=dimension, **opts)

    @property
    def data(self):
        return self.base.data

    def minimize(self, initial_guess=None, function_options=None, **opts):
        """
        Provides a uniform interface for minimization, basically just dispatching to the BaseSurface implementation if provided

        :param initial_guess: initial starting point for the minimization
        :type initial_guess: np.ndarray | None
        :param function_options:
        :type function_options: None | dict
        :param opts:
        :type opts:
        :return:
        :rtype:
        """

        # Some day I'll provide better minimization strategies for things like
        # Taylor series surfaces where there's a really well-defined analytic way
        # we can tackle the minimzation process
        # For now, though, we'll mostly just use scipy.optimize.minimize
        return self.base.minimize(initial_guess=initial_guess, function_options=function_options, **opts)

    @classmethod
    def detect_base(cls, data, opts):
        """
        Infers what type of base surface works for the data that's passed in.
        It's _super_ roughly done so...yeah generally better to pass the base class you want explicitly.
        But in the absence of that we can do this ?_?

        Basic strategy:
            1. look for options that go with specific methods
            2. look at data structures to guess
                i.   gradient as the first data arg + all data args are ndarrays -> Taylor Series
                ii.  callables as second arg -> Linear expansion or Linear fit
                iii. just like...one big array -> Interpolatin

        :param data:
        :type data: tuple
        :param opts:
        :type opts: dict
        :return:
        :rtype:
        """
        from ..FittableModels import LinearFitBasis

        # try a Taylor expansion
        if (
                any(key in opts for key in ['center', 'ref', 'weight_coefficients']) or (
                all(isinstance(d, np.ndarray) for d in data) and data[0].ndim==1
            )
        ):
            return TaylorSeriesSurface
        # next look into fitting
        elif (
                any(key in opts for key in ['basis']) or (
                len(data)==2 and (isinstance(data[1], LinearFitBasis) or all(callable(d) for d in data[1]))
            )
        ):
            # See if we were given fit coefficients or actual data to fit
            if data[0].ndim == 1:
                return LinearExpansionSurface
            else:
                return LinearFitSurface
        else:
            return InterpolatedSurface

    def __call__(self, gridpoints, **kwargs):
        return self.base(gridpoints, **kwargs)

    @property
    def center(self):
        return self.base.center
    @property
    def ref(self):
        return self.base.ref
    @property
    def expansion_tensors(self):
        return self.base.expansion_tensors

class MultiSurface:
    """
    A _reallly_ simple extension to the Surface infrastructure to handle vector valued functions,
    assuming each vector value corresponds to a different Surfaces
    """
    def __init__(self, *surfs):
        """

        :param surfs: a set of Surface objects to use when evaluating
        :type surfs: Iterable[Surface]
        """
        self.surfs = surfs
    def __call__(self, gridpoints, order=None, **kwargs):
        subvals = [s(gridpoints, order=order, **kwargs) for s in self.surfs]
        if order is None:
            return np.array(subvals)
        else:
            return [
                np.moveaxis(np.array([s[o] for s in subvals]), 0, -1)
                for o in range(order + 1)
            ]



import itertools
import numbers

__all__ = [
    "is_fixed_size",
    "consume",
    "chunked",
    "take_lists",
    "split",
    "split_by",
    "counts",
    "dict_diff",
    "transpose",
    "riffle",
    "flatten",
    "delete_duplicates"
]

def is_fixed_size(iterable):
    return hasattr(iterable, '__len__')

def consume(iterable, n, return_values=True):
    if return_values:
        return list(itertools.islice(iterable, n))
    else:
        return next(itertools.islice(iterable, n, n), None)

def chunked(a, upto):
    if is_fixed_size(a):
        l = len(a)
        blocks = (l // upto) +  (0 if (l % upto) == 0 else 1)
        for i in range(blocks):
            yield a[i*upto:(i+1)*upto]
    else:
        def consumer(a=a):
            return list(itertools.islice(a, upto))
        return iter(consumer, [])

def split(a, test=None):
    if is_fixed_size(a):
        prev = a[0]
        sep = 0
        for i, b in a[1]:
            if test(b, prev):
                yield a[sep:i+1]
                sep = i
                prev = b
    else:
        a, t = itertools.tee(a, 2)
        prev = next(t)
        sep = -1
        i = -1
        if test is None:
            test = lambda x,y: x != y
        for i, b in enumerate(t):
            if test(b, prev):
                yield consume(a, i + 1 - sep)
                sep = i
                prev = b
        yield consume(a, i + 1 - sep)

def split_by(a, canonicalizer):
    def test(next, prev, canonicalizer=canonicalizer, cache=[True, None]):
        if cache[0]:
            cache[1] = canonicalizer(prev)
        new = canonicalizer(next)
        ret = new != cache[1]
        cache[1] = new
        return ret
    return split(a, test)

def take_lists(a, splits):
    if is_fixed_size(a):
        sep = 0
        for s in splits:
            new = sep + s
            yield a[sep:new]
            sep = new
        if sep < len(a) - 1:
            yield a[sep:]
    else:
        for s in splits:
            yield consume(a, s)
def counts(iterable, test=None, hashable=True):
    if test is None: test = lambda x:x
    if hashable:
        counts_dict = {}
        for v in iterable:
            k = test(v)
            counts_dict[k] = counts_dict.get(k, 0) + 1
        return counts_dict
    else:
        keys = []
        counts = []
        for v in iterable:
            k = test(v)
            try:
                i = keys.index(k)
            except ValueError:
                keys.append(k)
                counts.append(1)
            else:
                counts[i] += 1

        return keys, counts

def dict_diff(iterable1:dict, iterable2):
    return {
        k:v
        for k,v in iterable1.items()
        if k not in iterable2 or v != iterable2[k]
    } | {k:iterable2[k] for k in iterable2.keys() - iterable1.keys()}

def transpose_iter(data, default=None):
    sentinel = object()
    any_full = True
    while any_full:
        any_full = False
        sublist = []
        for d in data:
            new = next(d, sentinel)
            if new is sentinel:
                sublist.append(default)
            else:
                any_full = True
                sublist.append(new)
        yield sublist
def transpose(data, default=None, pad=False):
    if is_fixed_size(data):
        fixed_data = is_fixed_size(data[0])
        if fixed_data:
            max_len = max(len(x) for x in data)
            if pad:
                return [
                    [d[i] if len(d) > i else default for d in data]
                    for i in range(max_len)
                ]
            else:
                return [
                    [d[i] for d in data if len(d) > i]
                    for i in range(max_len)
                ]
        else:
           return transpose_iter(data, default=default)
    else:
        return transpose(list(data), default=default) # need to cache results...

def riffle(a, b, *extras):
    if not is_fixed_size(a):
        a = list(a)
    l = len(a)
    n = -1
    for n,p in enumerate(zip(a, b, *extras)):
        if n == l - 1:
            yield p[0]
        else:
            for x in p:
                yield x
    for x in a[n+1:]: yield x # exhaust a

def _is_atomic(atomic_obj, atomic_types):
    if isinstance(atomic_obj, atomic_types): return True
    try:
        for _ in iter(atomic_obj): break
    except TypeError:
        return True
    return False

def flatten(iterable, atomic_types=None):
    if atomic_types is None:
        atomic_types = (int,str,float,numbers.Number)
    for o in iterable:
        if _is_atomic(o, atomic_types):
            yield o
        else:
            for f in flatten(o, atomic_types=atomic_types):
                yield f

def delete_duplicates(iterable, key=None, hashable=None, cache=None):
    hashable = hashable
    if cache is None:
        cache = set() if (hashable or hashable is None) else []
    elif hashable is None:
        hashable = hasattr(cache, 'add')
    if key is None:
        key = lambda x:x
    for o in iterable:
        test = key(o)
        if test in cache: continue
        if hashable is None:
            try:
                cache.add(test)
            except TypeError:
                cache = list(cache)
                hashable = False
        elif hashable:
            cache.add(test)
        else:
            cache.append(test)

        yield o



import abc
import functools
import os

import numpy as np
import uuid
from .NodeCommServer import NodeCommHandler, NodeCommClient

__all__ = [
    'EvaluationHandler',
    'EvaluationClient'
]

class EvaluationHandler(NodeCommHandler):
    @abc.abstractmethod
    def get_evaluators(self) -> 'dict[str,method]':
        ...

    def wrap_evaluator(self, name, evaluation_function):
        @functools.wraps(evaluation_function)
        def evaluate(args, kwargs):
            coords = args[0]
            use_file_io = isinstance(coords, str)
            if use_file_io:
                # we assume we got a .npz file of coords
                coords = np.load(coords)
            res = evaluation_function(coords, **kwargs)
            if use_file_io:
                #TODO: write this to temporary file
                id = str(uuid.uuid4())
                output = f'{name}-{id}'
                if isinstance(res, np.ndarray):
                    res = {'outfile':np.save(output+'.npy')}
                else:
                    res = {'outfile':np.save(output+'.npz')}
            else:
                if isinstance(res, np.ndarray):
                    res = {'result_array':res.tolist()}
                else:
                    res = {
                        k: v.tolist() if isinstance(v, np.ndarray) else v
                        for k, v in res.items()
                    }
            return res


        return evaluate

    def get_methods(self) -> 'dict[str,method]':
        return {
            k:self.wrap_evaluator(k,v)
            for k,v in self.get_evaluators().items()
        }

class EvaluationClient(NodeCommClient):
    def call(self, evaluator:str, coords:np.ndarray, filename=None, **kwargs):
        if filename is not None:
            coords = np.save(filename, coords)
        else:
            coords = np.asarray(coords).tolist()
        kwargs = {
            k:v.tolist() if isinstance(v, np.ndarray) else v
            for k,v in kwargs.items()
        }
        res = self.communicate(evaluator, (coords,), kwargs)
        if 'result_array' in res:
            res = res['result_array']
        else:
            if 'outfile' in res:
                oof = res['outfile']
                try:
                    res = np.load(oof)
                finally:
                    os.remove('outfile')
        return res


import enum
import numpy as np


__all__ = [
    "RotorTypes",
    "identify_rotor_type"
]

class RotorTypes(enum.Enum):
    Atom = "atom"
    Linear = "linear"
    Planar = "planar"
    Oblate = "oblate"
    Prolate = "prolate"
    Spherical = "spherical"
    Asymmetric = "asymmetric"

def identify_rotor_type(moms:np.ndarray, tol=1e-8, zero_tol=1e-6):
    diffs = np.diff(moms)
    eq_pos = np.where(diffs < tol)
    if len(eq_pos) > 0:
        pos = eq_pos[0]
    else:
        pos = []
    planar = np.abs((moms[0] + moms[1]) - moms[2]) < tol
    if np.min(moms) < zero_tol:
        if len(np.where(moms < zero_tol)[0]) == 2:
            type = RotorTypes.Atom
        else:
            type = RotorTypes.Linear
    elif len(pos) == 2:
        type = RotorTypes.Spherical
    elif len(pos) == 1:
        if pos[0] == 0:
            type = RotorTypes.Oblate
        else:
            type = RotorTypes.Prolate
    else:
        type = RotorTypes.Asymmetric

    return type, planar



import numpy as np
from collections import namedtuple
from .Parsers import ElectronicStructureLogReader
from ...Parsers import FileStreamReader, Number, RegexPattern, Integer, Word

__all__ = [
    "OrcaLogReader",
    "OrcaHessReader"
]


class OrcaLogReader(ElectronicStructureLogReader):
    components_name = "OrcaLogComponents"


class OrcaHessReader(FileStreamReader):

    def __init__(self, file, **kwargs):
        super().__init__(file, **kwargs)
        self._num_atoms = None

    ENumber = RegexPattern((Number, "E", Integer), "ENumber", dtype=float)
    matrix_types = {
        'normal_modes':ENumber,
    }
    array_types = {
        'vibrational_frequencies': Number,
        'dipole_derivatives': ENumber,
        'ir_spectrum': Number
    }
    @classmethod
    def get_special_handlers(cls):
        return {
            'atoms': cls.parse_atoms
            # 'ir_spectrum':cls.parse_ir
        }
    @classmethod
    def handle_orca_block(cls, tag, data:str):
        special_handler = cls.get_special_handlers().get(tag)
        if special_handler is not None:
            return special_handler(data)

        matrix_dtype = cls.matrix_types.get(tag)
        if matrix_dtype is not None:
            return cls.parse_matrix(data, data_pattern=matrix_dtype)

        array_type = cls.array_types.get(tag)
        if array_type is not None:
            return cls.parse_array(data, data_pattern=array_type)

        num_lines = data.count("\n")
        if num_lines == 1:
            return float(data.strip())

        header, rem = data.split("\n", 1)
        try:
            num_data = float(header)
        except TypeError:
            return data

        if num_data == num_lines - 1:
            return cls.parse_array(rem)
        else:
            return cls.parse_matrix(data)

    @classmethod
    def parse_matrix(cls, data, col_blocks=5, data_pattern=None):
        header, rem = data.split("\n", 1)
        header = header.split()
        if len(header) == 1:
            nrows = ncols = header[0]
        else:
            nrows, ncols = header
        del header

        nrows = int(nrows)
        ncols = int(ncols)
        if data_pattern is None:
            data_pattern = cls.ENumber
        nums = np.array(data_pattern.findall(data)).astype(float)
        if len(nums) != (nrows * ncols):
            raise ValueError(f"mismatch between number of elements {len(nums)} and matrix size ({nrows}x{ncols})")
        block_size = nrows * col_blocks
        nblocks = ncols // col_blocks
        arr = np.empty((nrows, ncols))
        for i in range(nblocks):
            start = block_size*i
            end = block_size*(i+1)
            col_offset = col_blocks * i
            arr[:, col_offset:col_offset + col_blocks] = nums[start:end].reshape(nrows, col_blocks)
        if ncols % col_blocks != 0:
            start = block_size*nblocks
            col_offset = ncols % col_blocks
            arr[:, -col_offset:] = nums[start:].reshape(nrows, col_offset)

        return arr

    @classmethod
    def parse_array(cls, data, data_pattern=None):
        header, rem = data.split("\n", 1)
        header = int(header.split()[0])
        if data_pattern is None:
            data_pattern = cls.ENumber
        nums = np.array(data_pattern.findall(data)).astype(float)
        res = nums.reshape(header, -1)
        if res.shape[1] == 1:
            res = res.flatten()
        return res

    OrcaCoords = namedtuple("OrcaCoords", ["atoms", "mass", "coords"])
    @classmethod
    def parse_atoms(cls, data):
        base_data = cls.parse_array(data, Number)
        atoms = Word.findall(data)
        return cls.OrcaCoords(
            atoms,
            base_data[:, 0],
            base_data[:, 1:]
        )

    def get_next_block(self):
        block = self.get_tagged_block("$", "\n\n")
        if block is None: return None
        tag, rem = block.split("\n", 1)
        return tag, rem

    def parse(self, tags=None, excludes=None):
        res = {}
        block = self.get_next_block()
        while block is not None:
            tag, data = block
            match = True
            if excludes is not None:
                match = tag not in tags
            if tags is not None:
                match = tag in tags
            if match:
                res[tag] = self.handle_orca_block(tag, data)
            block = self.get_next_block()

        return res






__all__ = [
    "RDMolecule"
]

import numpy as np, io, os
from .. import Numputils as nput
from ..Devutils import OutputRedirect

from .ChemToolkits import RDKitInterface
from .ExternalMolecule import ExternalMolecule

class RDMolecule(ExternalMolecule):
    """
    A simple interchange format for RDKit molecules
    """

    def __init__(self, rdconf, charge=None):
        #atoms, coords, bonds):
        self._rdmol = rdconf.GetOwningMol()
        super().__init__(rdconf)
        self.charge = charge

    @property
    def rdmol(self):
        if self._rdmol is None:
            self._rdmol = self.mol.GetOwningMol()
        return self._rdmol
    @property
    def atoms(self):
        mol = self.rdmol
        return [atom.GetSymbol() for atom in mol.GetAtoms()]
    @property
    def bonds(self):
        mol = self.rdmol
        return [
            [b.GetBeginAtomIdx(), b.GetEndAtomIdx(), b.GetBondTypeAsDouble()]
            for b in mol.GetBonds()
        ]
    @property
    def coords(self):
        return self.mol.GetPositions()
    @property
    def rings(self):
        return self.rdmol.GetRingInfo().AtomRings()
    @property
    def meta(self):
        return self.rdmol.GetPropsAsDict()

    def copy(self):
        Chem = self.chem_api()
        conf = self.mol
        new_mol = Chem.AddHs(Chem.Mol(self.rdmol), explicitOnly=True)
        new_mol.AddConformer(conf)
        return type(self).from_rdmol(new_mol,
                                     conf_id=conf.GetId(),
                                     charge=self.charge, sanitize=False,
                                     guess_bonds=False,
                                     add_implicit_hydrogens=False
                                     )
    @property
    def charges(self):
        from rdkit.Chem import AllChem
        AllChem.ComputeGasteigerCharges(self.rdmol)
        return [
            at.GetDoubleProp('_GasteigerCharge')
            for at in self.rdmol.GetAtoms()
        ]

    @classmethod
    def chem_api(cls):
        return RDKitInterface.submodule("Chem")
    @classmethod
    def from_rdmol(cls, rdmol, conf_id=0, charge=None, guess_bonds=False, sanitize=True,
                   add_implicit_hydrogens=False,
                   sanitize_ops=None):
        Chem = cls.chem_api() # to get nice errors
        rdmol = Chem.AddHs(rdmol, explicitOnly=not add_implicit_hydrogens)
        if guess_bonds:
            rdDetermineBonds = RDKitInterface.submodule("Chem.rdDetermineBonds")
            rdmol = Chem.Mol(rdmol)
            if charge is None:
                charge = 0
            rdDetermineBonds.DetermineConnectivity(rdmol, charge=charge)
            # return cls.from_rdmol(rdmol, conf_id=conf_id, guess_bonds=False, charge=charge)
        if sanitize:
            rdmolops = RDKitInterface.submodule("Chem.rdmolops")
            if sanitize_ops is None:
                sanitize_ops = (
                        rdmolops.SANITIZE_ALL
                        ^rdmolops.SANITIZE_PROPERTIES
                        # ^rdmolops.SANITIZE_ADJUSTHS
                        # ^rdmolops.SANITIZE_CLEANUP
                        ^rdmolops.SANITIZE_CLEANUP_ORGANOMETALLICS
                )
            rdmol = Chem.Mol(rdmol)
            Chem.SanitizeMol(rdmol, sanitize_ops)
        conf = rdmol.GetConformer(conf_id)
        return cls(conf, charge=charge)

    @classmethod
    def resolve_bond_type(cls, t):
        Chem = cls.chem_api()

        if abs(t - 1.5) < 1e-2:
            t = Chem.BondType.names["AROMATIC"]
        elif abs(t - 2.5) < 1e-2:
            t = Chem.BondType.names["TWOANDAHALF"]
        elif abs(t - 3.5) < 1e-2:
            t = Chem.BondType.names["TWOANDAHALF"]
        else:
            t = Chem.BondType.values[int(t)]

        return t
    @classmethod
    def from_coords(cls, atoms, coords, bonds=None, charge=None, guess_bonds=None):
        Chem = cls.chem_api()
        mol = Chem.EditableMol(Chem.Mol())
        mol.BeginBatchEdit()
        for a in atoms:
            a = Chem.Atom(a)
            mol.AddAtom(a)
        if bonds is not None:
            for b in bonds:
                if len(b) == 2:
                    i,j = b
                    t = 1
                else:
                    i,j,t = b
                if nput.is_numeric(t):
                    t = cls.resolve_bond_type(t)
                else:
                    t = Chem.BondType.names[t]
                mol.AddBond(int(i), int(j), t)
        mol.CommitBatchEdit()

        mol = mol.GetMol()
        mol = Chem.AddHs(mol, explicitOnly=True)
        conf = Chem.Conformer(len(atoms))
        conf.SetPositions(np.asanyarray(coords))
        conf.SetId(0)
        mol.AddConformer(conf)

        if guess_bonds is None:
            guess_bonds = bonds is None

        return cls.from_rdmol(mol, charge=charge, guess_bonds=guess_bonds)

    @classmethod
    def from_mol(cls, mol, coord_unit="Angstroms", guess_bonds=None):
        from ..Data import UnitsData

        return cls.from_coords(
            mol.atoms,
            mol.coords * UnitsData.convert(coord_unit, "Angstroms"),
            bonds=mol.bonds,
            charge=mol.charge,
            guess_bonds=guess_bonds
        )

    @classmethod
    def _load_sdf_conf(cls, stream, which=0):
        Chem = cls.chem_api()
        mol = None
        for i in range(which+1):
            mol = next(Chem.ForwardSDMolSupplier(stream, sanitize=False, removeHs=False))
        return mol
    @classmethod
    def from_sdf(cls, sdf_string, which=0):
        if os.path.isfile(sdf_string):
            with open(sdf_string, 'rb') as stream:
                mol = cls._load_sdf_conf(stream, which=which)
        else:
            mol = cls._load_sdf_conf(io.BytesIO(sdf_string.encode()), which=which)
        return cls.from_rdmol(mol)

    @classmethod
    def get_confgen_opts(cls):
        AllChem = cls.allchem_api()
        params = AllChem.ETKDGv3()
        params.useExpTorsionAnglePrefs = True
        params.useBasicKnowledge = True
        return params
    @classmethod
    def from_smiles(cls, smiles,
                    sanitize=False,
                    parse_name=True,
                    allow_cxsmiles=True,
                    strict_cxsmiles=True,
                    remove_hydrogens=False,
                    replacements=None,
                    add_implicit_hydrogens=False,
                    call_add_hydrogens=True,
                    num_confs=1, optimize=False, take_min=True,
                    force_field_type='mmff',
                    **opts):

        if os.path.isfile(smiles):
            with open(smiles) as f:
                smiles = f.read()
        Chem = cls.chem_api()

        params = Chem.SmilesParserParams()
        params.removeHs = remove_hydrogens
        params.sanitize = sanitize
        if replacements is not None:
            params.replacements = replacements
        params.parseName = parse_name
        params.allowCXSMILES = allow_cxsmiles
        params.strictCXSMILES = strict_cxsmiles
        for k,v in opts.items():
            setattr(params, k, v)

        rdkit_mol = Chem.MolFromSmiles(smiles, params)
        if not sanitize:
            rdkit_mol.UpdatePropertyCache()
        if call_add_hydrogens:
            mol = Chem.AddHs(rdkit_mol, explicitOnly=not add_implicit_hydrogens)
        else:
            mol = rdkit_mol

        return cls.from_base_mol(mol,
                                 num_confs=1, optimize=False, take_min=True,
                                 force_field_type='mmff'
                                 )

        # rdDistGeom = RDKitInterface.submodule("Chem.rdDistGeom")
        # rdDistGeom.EmbedMolecule(mol, num_confs, **cls.get_confgen_opts())

    @classmethod
    def from_base_mol(cls, mol,
                           conf_id=-1,
                           num_confs=1,
                           optimize=False,
                           take_min=None,
                           force_field_type='mmff',
                           **mol_opts):
        try:
            conf = mol.GetConformer(conf_id)
        except ValueError:
            conf = None
        if conf:
            return cls.from_rdmol(mol, conf_id, **mol_opts)
        else:
            if conf_id > num_confs + 1:
                num_confs = conf_id
            return cls.from_no_conformer_molecule(mol,
                                                  num_confs=num_confs,
                                                  optimize=optimize,
                                                  take_min=conf_id < 0 if take_min is None else take_min,
                                                  force_field_type=force_field_type,
                                                  **mol_opts
                                                  )

    @classmethod
    def from_no_conformer_molecule(cls,
                                   mol,
                                   num_confs=1,
                                   optimize=False,
                                   take_min=True,
                                   force_field_type='mmff',
                                   add_implicit_hydrogens=False,
                                   **etc
                                   ):
        AllChem = cls.allchem_api()

        AllChem.AddHs(mol, explicitOnly=not add_implicit_hydrogens)

        with OutputRedirect():
            conformer_set = AllChem.EmbedMultipleConfs(mol, numConfs=num_confs, params=cls.get_confgen_opts())
        if optimize:
            rdForceFieldHelpers = RDKitInterface.submodule("Chem.rdForceFieldHelpers")
            if force_field_type == 'mmff':
                rdForceFieldHelpers.MMFFOptimizeMoleculeConfs(mol)
            elif force_field_type == 'uff':
                rdForceFieldHelpers.UFFOptimizeMoleculeConfs(mol)
            else:
                raise NotImplementedError(f"no basic preoptimization support for {force_field_type}")

        if take_min and num_confs > 1:
            conf_ids = list(conformer_set)
            force_field_type = cls.get_force_field_type(force_field_type)
            if isinstance(force_field_type, (list, tuple)):
                force_field_type, prop_gen = force_field_type
            else:
                prop_gen = None

            if prop_gen is not None:
                props = prop_gen(mol)
            else:
                props = None

            engs = [
                force_field_type(mol, props, confId=conf_id).CalcEnergy()
                for conf_id in conf_ids
            ]

            conf_id = conf_ids[np.argmin(engs)]
        else:
            conf_id = 0

        return cls.from_rdmol(mol, conf_id=conf_id, add_implicit_hydrogens=add_implicit_hydrogens, **etc)

    def to_smiles(self):
        return self.chem_api().MolToSmiles(self.rdmol)

    @classmethod
    def from_molblock(cls,
                      molblock,
                      add_implicit_hydrogens=False,
                      call_add_hydrogens=True,
                      ):
        Chem = cls.chem_api()
        if os.path.isfile(molblock):
            mol = Chem.MolFromMolFile(molblock, sanitize=False, removeHs=False)
        else:
            mol = Chem.MolFromMolBlock(molblock, sanitize=False, removeHs=False)
        return cls.from_rdmol(mol, add_implicit_hydrogens=add_implicit_hydrogens)

    @classmethod
    def allchem_api(cls):
        return RDKitInterface.submodule("Chem.AllChem")
    @classmethod
    def get_force_field_type(cls, ff_type):
        AllChem = cls.allchem_api()

        if isinstance(ff_type, str):
            if ff_type == 'mmff':
                ff_type = (AllChem.MMFFGetMoleculeForceField, AllChem.MMFFGetMoleculeProperties)
            elif ff_type == 'uff':
                ff_type = (AllChem.UFFGetMoleculeForceField, None)
            else:
                raise ValueError(f"can't get RDKit force field type from '{ff_type}")

        return ff_type

    def get_force_field(self, force_field_type='mmff', conf=None, mol=None, conf_id=None, **extra_props):
        if conf is None:
            if mol is None:
                mol = self
            if conf_id is None:
                conf_id = mol.mol.GetId()
            mol = mol.rdmol
        else:
            if mol is None:
                mol = conf.GetOwningMol()
            if conf_id is None:
                conf_id = conf.GetId()
            if np.sum(np.abs(conf.GetPositions() - mol.GetConformer(conf_id).GetPositions())) > 1e-6:
                print(mol)
                print(self.rdmol)
                raise ValueError(
                    conf, mol.GetConformer(conf_id)
                )

        force_field_type = self.get_force_field_type(force_field_type)
        if isinstance(force_field_type, (list, tuple)):
            force_field_type, prop_gen = force_field_type
        else:
            prop_gen = None

        if prop_gen is not None:
            props = prop_gen(mol)
        else:
            props = None

        if props is not None:
            return force_field_type(mol, props, confId=conf_id, **extra_props)
        else:
            return force_field_type(mol, confId=conf_id, **extra_props)

    def evaluate_charges(self, coords, model='gasteiger'):
        if model == 'gasteiger':
            from rdkit.Chem import AllChem
            AllChem.ComputeGasteigerCharges(self.rdmol)
            return [
                at.GetDoubleProp('_GasteigerCharge')
                for at in self.rdmol.GetAtoms()
            ]
        else:
            raise ValueError(f"charge model {model} not supported in RDKit")

    def calculate_energy(self, geoms=None, force_field_generator=None, force_field_type='mmff', conf_id=None):
        Chem = self.chem_api()
        if conf_id is None:
            conf_id = self.mol.GetId()
        mol = Chem.Mol(self.rdmol, confId=conf_id)
        conf = mol.GetConformer(0)

        if force_field_generator is None:
            force_field_generator = self.get_force_field
        if geoms is not None:
            cur_geom = np.array(self.mol.GetPositions()).reshape(-1, 3)
            geoms = np.asanyarray(geoms)
            base_shape = geoms.shape[:-2]
            geoms = geoms.reshape((-1,) + cur_geom.shape)
            vals = np.empty(len(geoms), dtype=float)
            try:
                for i,g in enumerate(geoms):
                    conf.SetPositions(g.copy())
                    ff = force_field_generator(force_field_type, conf=conf, mol=mol)
                    vals[i] = ff.CalcEnergy()
            finally:
                conf.SetPositions(cur_geom)
            return vals.reshape(base_shape)
        else:
            ff = force_field_generator(force_field_type, conf_id=conf_id)
            return ff.CalcEnergy()

    def calculate_gradient(self, geoms=None, force_field_generator=None, force_field_type='mmff', conf_id=None):
        if force_field_generator is None:
            force_field_generator = self.get_force_field

        Chem = self.chem_api()
        if conf_id is None:
            conf_id = self.mol.GetId()
        mol = Chem.Mol(self.rdmol, confId=conf_id)
        conf = mol.GetConformer(0)

        cur_geom = np.array(conf.GetPositions()).reshape(-1, 3)
        if geoms is not None:
            geoms = np.asanyarray(geoms)
            base_shape = geoms.shape[:-2]
            geoms = geoms.reshape((-1,) + cur_geom.shape)
            vals = np.empty((len(geoms), np.prod(cur_geom.shape, dtype=int)), dtype=float)
            try:
                for i, g in enumerate(geoms):
                    new_geom = g.copy().view(np.ndarray)
                    conf.SetPositions(new_geom)
                    ff = force_field_generator(force_field_type, conf=conf, mol=mol)
                    vals[i] = ff.CalcGrad()
            finally:
                conf.SetPositions(cur_geom)
            return vals.reshape(base_shape + (-1,))
        else:
            ff = force_field_generator(force_field_type, conf_id=conf_id, conf=conf, mol=mol)
            return np.array(ff.CalcGrad()).reshape(-1)

    def calculate_hessian(self, force_field_generator=None, force_field_type='mmff', stencil=5, mesh_spacing=.01, **fd_opts):
        from ..Zachary import FiniteDifferenceDerivative

        cur_geom = np.array(self.mol.GetPositions()).reshape(-1, 3)

        # if force_field_generator is None:
        #     force_field_generator = self.get_force_field(force_field_type)

        # def eng(structs):
        #     structs = structs.reshape(structs.shape[:-1] + (-1, 3))
        #     new_grad = self.calculate_energy(structs, ff=ff)
        #     return new_grad
        # der = FiniteDifferenceDerivative(eng, function_shape=((0,), (0,)), stencil=stencil, mesh_spacing=mesh_spacing, **fd_opts)
        # return der.derivatives(cur_geom.flatten()).derivative_tensor(2)

        def jac(structs):
            structs = structs.reshape(structs.shape[:-1] + (-1, 3))
            new_grad = self.calculate_gradient(structs,
                                               force_field_generator=force_field_generator,
                                               force_field_type=force_field_type
                                               )
            return new_grad
        der = FiniteDifferenceDerivative(jac, function_shape=((0,), (0,)), stencil=stencil, mesh_spacing=mesh_spacing, **fd_opts)
        return der.derivatives(cur_geom.flatten()).derivative_tensor(1)

    def get_optimizer_params(self, maxAttempts=1000, useExpTorsionAnglePrefs=True, useBasicKnowledge=True, **etc):
        AllChem = self.allchem_api()

        params = AllChem.ETKDGv3()
        params.maxAttempts = maxAttempts  # Increase the number of attempts
        params.useExpTorsionAnglePrefs = useExpTorsionAnglePrefs
        params.useBasicKnowledge = useBasicKnowledge
        for k,v in etc.items():
            setattr(params, k, v)

        return params

    def optimize_structure(self, geoms=None, force_field_type='mmff', optimizer=None, maxIters=1000, **opts):

        if optimizer is None:
            ff_helpers = RDKitInterface.submodule("Chem.rdForceFieldHelpers")
            def optimizer(mol, **etc):
                ff = mol.get_force_field(force_field_type)
                return ff_helpers.OptimizeMolecule(ff)

        maxIters = int(maxIters)
        if geoms is not None:
            cur_geom = np.array(self.mol.GetPositions()).reshape(-1, 3)
            geoms = np.asanyarray(geoms, dtype=float)
            base_shape = geoms.shape[:-2]
            geoms = geoms.reshape((-1,) + cur_geom.shape)
            opt_vals = np.empty((len(geoms),), dtype=int)
            opt_geoms = np.empty_like(geoms)
            try:
                for i, g in enumerate(geoms):
                    self.mol.SetPositions(g)
                    opt_vals[i] = optimizer(self, maxIters=maxIters, **opts)
                    opt_geoms[i] = self.mol.GetPositions()
            finally:
                self.mol.SetPositions(cur_geom)
            return opt_vals.reshape(base_shape), opt_geoms.reshape(base_shape + opt_geoms.shape[1:]), {}
        else:
            opt = optimizer(self, maxIters=maxIters, **opts)
            return opt, self.mol.GetPositions(), {}

    def show(self):
        return RDKitInterface.submodule('Chem.Draw.IPythonConsole').drawMol3D(
            self.mol.GetOwningMol(),
            confId=self.mol.GetId()
            # view=None, confId=-1, drawAs=None, bgColor=None, size=None
        )




from ..Data import AtomData
import abc, typing

__all__ = [
    "ExternalMolecule"
]


class ExternalMolecule(metaclass=abc.ABCMeta):
    """
    Defines a common interface so that a programs can define one interface and not need to
    adjust for every new type of chemistry package they want to support.
    Not all properties need to be well-defined on every type of molecule, but this defines the core interface
    that _must_ be implemented
    """

    def __init__(self, external_mol):
        self.mol = external_mol

    @property
    @abc.abstractmethod
    def atoms(self):
        ...
    @property
    @abc.abstractmethod
    def coords(self):
        ...
    @property
    def masses(self):
        return [AtomData[x, "Mass"] for x in self.atoms]

    @property
    def bonds(self):
        raise NotImplementedError("{} doesn't implement {}".format(
            type(self).__name__,
            "bonds"
        ))

    @property
    def charges(self):
        raise NotImplementedError("{} doesn't implement {}".format(
            type(self).__name__,
            "charges"
        ))

    @classmethod
    @abc.abstractmethod
    def from_coords(cls, atoms, coords, **etc) -> 'typing.Self':
        ...

    @classmethod
    def from_mol(cls, atoms, coords, **etc) -> 'typing.Self':
        ...

    def show(self):
        raise NotImplementedError("{} doesn't implement {}".format(
            type(self).__name__,
            "show"
        ))




__all__ = [
    "merge_sets"
]


def merge_sets(iterable):
    groups:list[set] = []
    for s in iterable:
        s = set(s)
        for g in groups:
            if len(s.intersection(g)) > 0:
                g.update(s)
                break
        else:
            groups.append(s)

    skipped = [False]*len(groups)
    final = []
    for n,g in enumerate(groups):
        if skipped[n]: continue
        for m,g1 in enumerate(groups[n:]):
            j = n + m
            if skipped[j]: continue
            if len(g1.intersection(g)) > 0:
                g.update(g1)
                skipped[j] = True
        final.append(g)
    return final




import io

import numpy as np
from .FileStreamer import FileStreamReader, FileStreamerTag
# from .RegexPatterns import Word, Number

__all__ = [
    "XYZParser"
]

class XYZParser(FileStreamReader):
    def __init__(self, file, has_comments=True, **kw):
        self.has_comments = has_comments
        super().__init__(file, **kw)
    def find_block(self):
        int_tag = self.get_tagged_block(None, '\n')
        if int_tag is None: return None
        num_follows = int(int_tag)
        if not self.has_comments:
            num_follows = num_follows - 1
        full_tag = FileStreamerTag('\n', follow_ups=['\n']*num_follows)
        return self.get_tagged_block(None, full_tag, allow_terminal=True)

    def parse_xyz_block(self, block, include_comment=True):
        if self.has_comments:
            comment, block = block.split('\n', 1)
        else:
            comment = None
        atoms = np.loadtxt(io.StringIO(block), usecols=[0], dtype=str)
        coords = np.loadtxt(io.StringIO(block), usecols=[1, 2, 3])

        if include_comment:
            return comment, atoms, coords
        else:
            return atoms, coords

    def parse(self, max_blocks=None, include_comment=True):
        blocks = []
        block = self.find_block()
        if max_blocks is not None:
            for i in range(max_blocks-1):
                blocks.append(self.parse_xyz_block(block, include_comment=include_comment))
                block = self.find_block()
                if block is None:
                    break
        else:
            while block is not None:
                blocks.append(self.parse_xyz_block(block, include_comment=include_comment))
                block = self.find_block()

        return blocks




"""
A plotting framework that builds off of `matplotlib`, but potentially could use a different backend.
The design is intended to mirror the `Graphics` framework in Mathematica and where possible option
names have been chosen to be the same.
Difficulties with `matplotlib` prevent a perfect mirror but the design is consistent.
There are a few primary divisions:
1. `Graphics`/`Graphics3D`/`GraphicsGrid` provide basic access to `matplotlib.figure` and `matplotlib.axes`
    they also hold a `GraphicsPropertyManager`/`GraphicsPropertyManager3D` that manages all properties
    (`image_size`, `axes_label`, `ticks_style`, etc.).
    The full lists can be found on the relevant reference pages and are bound as `properties` on the
    `Graphics`/`Graphics3D` instances.
2. `Plot/Plot3D` and everything in the `Plots` subpackage provide concrete instances of common plots
    with nice names/consistent with Mathematica for discoverability but primarily fall back onto
    `matplotlib` built-in methods and then allow for restyling/data reuse, etc.
3. `Primitives` provide direct access to the shapes that are actually plotted on screen (i.e. `matplotlib.Patch` objects)
    in a convenient way to add on to existing plots
4. `Styling` provides access to theme management/construction

Image/animation support and other back end support for 3D graphics (`VTK`) are provided at the experimental level.
"""

__all__ = []
from .Graphics import *; from .Graphics import __all__ as exposed
__all__ += exposed
from .Plots import *; from .Plots import __all__ as exposed
__all__ += exposed
from .Primitives import *; from .Primitives import __all__ as exposed
__all__ += exposed
from .Interactive import *; from .Interactive import __all__ as exposed
__all__ += exposed
from .Styling import *; from .Styling import __all__ as exposed
__all__ += exposed
from .Image import *; from .Image import __all__ as exposed
__all__ += exposed
from .Properties import *; from .Properties import __all__ as exposed
__all__ += exposed
from .SceneJSON import *; from .SceneJSON import __all__ as exposed
__all__ += exposed
from .X3DInterface import *; from .X3DInterface import __all__ as exposed
__all__ += exposed
from .Colors import *; from .Colors import __all__ as exposed
__all__ += exposed
